"""Checks specific to intracellular electrophysiology neurodata types."""

from typing import Optional

from pynwb.icephys import IntracellularElectrode

from .._registration import Importance, InspectorMessage, register_check


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=IntracellularElectrode)
def check_intracellular_electrode_cell_id_exists(
    intracellular_electrode: IntracellularElectrode,
) -> Optional[InspectorMessage]:
    """
    Check if the IntracellularElectrode contains a cell_id.

    Best Practice: TODO
    """
    if hasattr(intracellular_electrode, "cell_id") and intracellular_electrode.cell_id is None:
        return InspectorMessage(message="Please include a unique cell_id associated with this IntracellularElectrode.")

    return None



"""Helper functions for internal use across the testing suite."""

import os
from datetime import datetime
from pathlib import Path
from typing import Optional
from urllib import request
from uuid import uuid4

import zarr
from hdmf.backends.hdf5 import HDF5IO
from hdmf.backends.io import HDMFIO
from packaging.version import Version
from pynwb import NWBHDF5IO, NWBFile
from pynwb.image import ImageSeries

from ..utils import get_package_version, is_module_installed, strtobool

TESTING_FILES_FOLDER_PATH = os.environ.get("TESTING_FILES_FOLDER_PATH", None)


def check_streaming_tests_enabled() -> tuple[bool, Optional[str]]:
    """
    General purpose helper for determining if the testing environment can support S3 DANDI streaming.

    Returns the boolean status of the check and, if False, provides a string reason for the failure for the user to
    utilize as they please (raise an error or warning with that message, print it, or ignore it).
    """
    failure_reason: str = ""

    environment_skip_flag = os.environ.get("NWBI_SKIP_NETWORK_TESTS", "")
    environment_skip_flag_bool = (
        strtobool(os.environ.get("NWBI_SKIP_NETWORK_TESTS", "")) if environment_skip_flag != "" else False
    )
    if environment_skip_flag_bool:
        failure_reason += "Environmental variable set to skip network tests."

    streaming_enabled, streaming_failure_reason = check_streaming_enabled()
    if not streaming_enabled and streaming_failure_reason is not None:
        failure_reason += streaming_failure_reason

    have_dandi = is_module_installed("dandi")
    if not have_dandi:
        failure_reason += "The DANDI package is not installed on the system."

    have_remfile = is_module_installed("remfile")
    if not have_remfile:
        failure_reason += "The `remfile` package is not installed on the system."

    return_failure_reason: Optional[str] = None
    if failure_reason != "":
        return_failure_reason = failure_reason

    return streaming_enabled and not environment_skip_flag_bool and have_dandi, return_failure_reason


def generate_testing_files() -> None:  # pragma: no cover
    """Generate a local copy of the NWB files required for all tests."""
    generate_image_series_testing_files()


def generate_image_series_testing_files() -> None:  # pragma: no cover
    """Generate a local copy of the NWB files required for the image series tests."""
    assert get_package_version(name="pynwb") == Version("2.1.0"), "Generating the testing files requires PyNWB v2.1.0!"
    assert TESTING_FILES_FOLDER_PATH is not None, "The `TESTING_FILES_FOLDER_PATH` environment variable is not set!"

    testing_folder = Path(TESTING_FILES_FOLDER_PATH)
    testing_folder.mkdir(exist_ok=True)

    movie_1_file_path = testing_folder / "temp_movie_1.mov"
    nested_folder = testing_folder / "nested_folder"
    nested_folder.mkdir(exist_ok=True)
    movie_2_file_path = nested_folder / "temp_movie_2.avi"
    for file_path in [movie_1_file_path, movie_2_file_path]:
        with open(file=file_path, mode="w") as file:
            file.write("Not a movie file, but at least it exists.")
    nwbfile = make_minimal_nwbfile()
    nwbfile.add_acquisition(
        ImageSeries(
            name="TestImageSeriesGoodExternalPaths",
            rate=1.0,
            external_file=[
                "/".join([".", movie_1_file_path.name]),
                "/".join([".", nested_folder.name, movie_2_file_path.name]),
            ],
        )
    )
    nwbfile.add_acquisition(
        ImageSeries(
            name="TestImageSeriesExternalPathDoesNotExist",
            rate=1.0,
            external_file=["madeup_file.mp4"],
        )
    )
    absolute_file_path = str(Path("madeup_file.mp4").absolute())
    nwbfile.add_acquisition(
        ImageSeries(name="TestImageSeriesExternalPathIsNotRelative", rate=1.0, external_file=[absolute_file_path])
    )
    with NWBHDF5IO(path=testing_folder / "image_series_testing_file.nwb", mode="w") as io:
        io.write(nwbfile)


def make_minimal_nwbfile() -> NWBFile:
    """
    Most basic NWBFile that can exist.

    TODO: replace with pynwb.mock if we can require minimal PyNWB version (or perhaps just for a testing)
    """
    return NWBFile(session_description="", identifier=str(uuid4()), session_start_time=datetime.now().astimezone())


def check_streaming_enabled() -> tuple[bool, Optional[str]]:
    """
    General purpose helper for determining if the environment can support S3 DANDI streaming.

    Returns the boolean status of the check and, if False, provides a string reason for the failure for the user to
    utilize as they please (raise an error or warning with that message, print it, or ignore it).
    """
    try:
        request.urlopen("https://dandiarchive.s3.amazonaws.com/ros3test.nwb", timeout=1)
    except request.URLError:  # type: ignore
        return False, "Internet access to DANDI failed."
    return True, None


def check_hdf5_io_open(io: HDF5IO) -> bool:
    """
    Check if an h5py.File object is open by using the file .id attribute, which is invalid when the file is closed.
    """
    return io._file.id.valid


def check_zarr_io_open(io: HDMFIO) -> bool:
    """For Zarr, the private attribute `_ZarrIO__file` is set to a `zarr.group` on open."""
    return isinstance(io._ZarrIO__file, zarr.Group)



from ._behavior import (
    check_compass_direction_unit,
    check_spatial_series_degrees_magnitude,
    check_spatial_series_dims,
    check_spatial_series_radians_magnitude,
)
from ._ecephys import (
    check_ascending_spike_times,
    check_electrical_series_dims,
    check_electrical_series_reference_electrodes_table,
    check_negative_spike_times,
    check_spike_times_not_in_unobserved_interval,
)
from ._general import (
    check_description,
    check_name_colons,
    check_name_slashes,
)
from ._icephys import (
    check_intracellular_electrode_cell_id_exists,
)
from ._image_series import (
    check_image_series_data_size,
    check_image_series_external_file_relative,
    check_image_series_external_file_valid,
)
from ._images import (
    check_index_series_points_to_image,
    check_order_of_images_len,
    check_order_of_images_unique,
)
from ._nwb_containers import (
    check_empty_string_for_optional_attribute,
    check_large_dataset_compression,
    check_small_dataset_compression,
)
from ._nwbfile_metadata import (
    check_doi_publications,
    check_experiment_description,
    check_experimenter_exists,
    check_experimenter_form,
    check_institution,
    check_keywords,
    check_processing_module_name,
    check_session_id_no_slashes,
    check_session_start_time_future_date,
    check_session_start_time_old_date,
    check_subject_age,
    check_subject_exists,
    check_subject_id_exists,
    check_subject_id_no_slashes,
    check_subject_proper_age_range,
    check_subject_sex,
    check_subject_species_exists,
    check_subject_species_form,
)
from ._ogen import (
    check_optogenetic_stimulus_site_has_optogenetic_series,
)
from ._ophys import (
    check_emission_lambda_in_nm,
    check_excitation_lambda_in_nm,
    check_plane_segmentation_image_mask_shape_against_ref_images,
    check_roi_response_series_dims,
    check_roi_response_series_link_to_plane_segmentation,
)
from ._tables import (
    check_col_not_nan,
    check_column_binary_capability,
    check_dynamic_table_region_data_validity,
    check_empty_table,
    check_ids_unique,
    check_single_row,
    check_table_time_columns_are_not_negative,
    check_table_values_for_dict,
    check_time_interval_time_columns,
    check_time_intervals_stop_after_start,
)
from ._time_series import (
    check_data_orientation,
    check_missing_unit,
    check_rate_is_not_zero,
    check_rate_is_positive,
    check_regular_timestamps,
    check_resolution,
    check_timestamp_of_the_first_sample_is_not_negative,
    check_timestamps_ascending,
    check_timestamps_match_first_dimension,
    check_timestamps_without_nans,
)

__all__ = [
    "check_negative_spike_times",
    "check_electrical_series_dims",
    "check_electrical_series_reference_electrodes_table",
    "check_spike_times_not_in_unobserved_interval",
    "check_description",
    "check_name_slashes",
    "check_name_colons",
    "check_image_series_data_size",
    "check_image_series_external_file_relative",
    "check_image_series_external_file_valid",
    "check_order_of_images_unique",
    "check_order_of_images_len",
    "check_index_series_points_to_image",
    "check_empty_string_for_optional_attribute",
    "check_small_dataset_compression",
    "check_large_dataset_compression",
    "check_keywords",
    "check_institution",
    "check_subject_age",
    "check_subject_sex",
    "check_subject_exists",
    "check_doi_publications",
    "check_experimenter_form",
    "check_experimenter_exists",
    "check_experiment_description",
    "check_subject_id_exists",
    "check_subject_id_no_slashes",
    "check_subject_species_exists",
    "check_subject_species_form",
    "check_subject_proper_age_range",
    "check_session_id_no_slashes",
    "check_session_start_time_future_date",
    "check_processing_module_name",
    "check_session_start_time_old_date",
    "check_optogenetic_stimulus_site_has_optogenetic_series",
    "check_excitation_lambda_in_nm",
    "check_plane_segmentation_image_mask_shape_against_ref_images",
    "check_roi_response_series_dims",
    "check_emission_lambda_in_nm",
    "check_roi_response_series_link_to_plane_segmentation",
    "check_single_row",
    "check_ids_unique",
    "check_empty_table",
    "check_col_not_nan",
    "check_column_binary_capability",
    "check_dynamic_table_region_data_validity",
    "check_time_interval_time_columns",
    "check_time_intervals_stop_after_start",
    "check_table_values_for_dict",
    "check_table_time_columns_are_not_negative",
    "check_resolution",
    "check_missing_unit",
    "check_regular_timestamps",
    "check_timestamps_ascending",
    "check_data_orientation",
    "check_timestamps_without_nans",
    "check_timestamps_match_first_dimension",
    "check_timestamp_of_the_first_sample_is_not_negative",
    "check_rate_is_not_zero",
    "check_intracellular_electrode_cell_id_exists",
    "check_compass_direction_unit",
    "check_spatial_series_radians_magnitude",
    "check_spatial_series_dims",
    "check_spatial_series_degrees_magnitude",
    "check_ascending_spike_times",
    "check_rate_is_positive",
]



"""Checks for types belonging to the pynwb.behavior module."""

from typing import Iterable, Optional

import numpy as np
from pynwb.behavior import CompassDirection, SpatialSeries

from .._registration import Importance, InspectorMessage, register_check


@register_check(importance=Importance.CRITICAL, neurodata_type=SpatialSeries)
def check_spatial_series_dims(spatial_series: SpatialSeries) -> Optional[InspectorMessage]:
    """
    Check if a SpatialSeries has the correct dimensions.

    Best Practice: :ref:`best_practice_spatial_series_dimensionality`
    """
    if len(spatial_series.data.shape) > 1 and spatial_series.data.shape[1] > 3:
        return InspectorMessage(
            message="SpatialSeries should have 1 column (x), 2 columns (x, y), or 3 columns (x, y, z)."
        )

    return None


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=CompassDirection)
def check_compass_direction_unit(compass_direction: CompassDirection) -> Optional[Iterable[InspectorMessage]]:
    """
    Check that SpatialSeries objects within CompassDirection objects have units 'degrees' or 'radians'.

    Best Practice: :ref:`best_practice_spatial_series_units`
    """
    for spatial_series in compass_direction.spatial_series.values():
        if spatial_series.unit not in ("degrees", "radians"):
            yield InspectorMessage(
                message=f"SpatialSeries objects inside a CompassDirection object should be angular and should have a "
                f"unit of 'degrees' or 'radians', but '{spatial_series.name}' has units '{spatial_series.unit}'."
            )

    return None


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=SpatialSeries)
def check_spatial_series_radians_magnitude(
    spatial_series: SpatialSeries, nelems: int = 200
) -> Optional[InspectorMessage]:
    """
    Check that SpatialSeries with units radians have data values between -2*pi and 2*pi.

    Best Practice: :ref:`best_practice_spatial_series_values`
    """
    if spatial_series.unit in ("radian", "radians"):
        data = spatial_series.data[:nelems]
        if np.any(data > (2 * np.pi)) or np.any(data < (-2 * np.pi)):
            return InspectorMessage(
                message="SpatialSeries with units of radians must have values between -2pi and 2pi."
            )

    return None


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=SpatialSeries)
def check_spatial_series_degrees_magnitude(
    spatial_series: SpatialSeries, nelems: int = 200
) -> Optional[InspectorMessage]:
    """
    Check that SpatialSeries with units of degrees have data values between -360 and 360.

    Best Practice: :ref:`best_practice_spatial_series_values`
    """
    if spatial_series.unit in ("degree", "degrees"):
        data = spatial_series.data[:nelems]
        if np.any(data > 360) or np.any(data < -360):
            return InspectorMessage(
                message="SpatialSeries with units of degrees must have values between -360 and 360."
            )

    return None



"""Check functions that examine any general neurodata_type with the available attributes."""

from typing import Optional

from .._registration import Importance, InspectorMessage, register_check

COMMON_DESCRIPTION_PLACEHOLDERS = ["no description", "no desc", "none", "placeholder"]


@register_check(importance=Importance.CRITICAL, neurodata_type=None)
def check_name_slashes(neurodata_object: object) -> Optional[InspectorMessage]:
    """Check if there  has been added for the session."""
    if hasattr(neurodata_object, "name") and any((x in neurodata_object.name for x in ["/", "\\"])):
        return InspectorMessage(message="Object name contains slashes.")

    return None


@register_check(importance=Importance.BEST_PRACTICE_SUGGESTION, neurodata_type=None)
def check_name_colons(neurodata_object: object) -> Optional[InspectorMessage]:
    """Check if an object name contains a colon."""
    if hasattr(neurodata_object, "name") and ":" in neurodata_object.name:
        return InspectorMessage(message="Object name contains colons.")

    return None


@register_check(importance=Importance.BEST_PRACTICE_SUGGESTION, neurodata_type=None)
def check_description(neurodata_object: object) -> Optional[InspectorMessage]:
    """
    Check if the description is a not missing or a placeholder.

    Best Practice: :ref:`best_practice_placeholders`
    """
    if not hasattr(neurodata_object, "description"):
        return None

    description = neurodata_object.description
    if description is not None and type(description) is not str:
        return None
    if description is None or description.strip(" ") == "":
        return InspectorMessage(message="Description is missing.")
    if description.lower().strip(".") in COMMON_DESCRIPTION_PLACEHOLDERS:
        return InspectorMessage(message=f"Description ('{description}') is a placeholder.")

    return None



"""Primary functions for inspecting NWBFiles."""

import importlib
import importlib.metadata
import json
import os
from pathlib import Path
from typing import Union

import click

from ._configuration import load_config
from ._dandi_inspection import inspect_dandi_file_path, inspect_dandiset, inspect_url
from ._formatting import (
    InspectorOutputJSONEncoder,
    _get_report_header,
    format_messages,
    print_to_console,
    save_report,
)
from ._nwb_inspection import inspect_all
from ._types import Importance
from .utils import strtobool


@click.command()
@click.argument("path", nargs=1)
@click.option(
    "--stream",
    help="Whether or not to stream file content from the DANDI archive. Automatically enabled if path is a URL.",
    is_flag=True,
    required=False,
    default=False,
)
@click.option(
    "--version-id",
    help="The version ID of the Dandiset to inspect. Defaults to 'draft' if '--stream' is specified.",
    type=str,
    required=False,
    default=None,
)
@click.option(
    "--report-file-path",
    default=None,
    help="Save path for the report file.",
    type=click.Path(writable=True),
)
@click.option(
    "--config",
    help="Name of internal config or path to a custom YAML file.",
    type=str,
    required=False,
    default=None,
)
@click.option("--levels", help="Comma-separated names of InspectorMessage attributes to organize by.")
@click.option(
    "--reverse", help="Comma-separated booleans corresponding to reversing the order for each value of 'levels'."
)
@click.option("--overwrite", help="Overwrite an existing report file at the location.", is_flag=True)
@click.option("--ignore", help="Comma-separated names of checks to skip.")
@click.option("--select", help="Comma-separated names of checks to run.")
@click.option(
    "--threshold",
    default="BEST_PRACTICE_SUGGESTION",
    type=click.Choice(["CRITICAL", "BEST_PRACTICE_VIOLATION", "BEST_PRACTICE_SUGGESTION"]),
    help="Ignores tests with an assigned importance below this threshold.",
)
@click.option("--json-file-path", help="Write json output to this location.")
@click.option("--n-jobs", help="Number of jobs to use in parallel.", default=1)
@click.option("--skip-validate", help="Skip the PyNWB validation step.", is_flag=True)
@click.option(
    "--detailed",
    help=(
        "If file_path is the last of 'levels' (the default), identical checks will be aggregated in the display. "
        "Use '--detailed' to see the complete report."
    ),
    is_flag=True,
)
@click.option("--progress-bar", help="Set this flag to False to disable display of the progress bar.")
@click.option(
    "--modules",
    help="Modules to import prior to reading the file(s). Necessary for registration of custom checks functions.",
    type=str,
    required=False,
    default=None,
)
@click.version_option(version=importlib.metadata.version(distribution_name="nwbinspector"))
def _nwbinspector_cli(
    *,
    path: str,
    stream: bool = False,
    version_id: Union[str, None] = None,
    report_file_path: Union[str, None] = None,
    config: Union[str, None] = None,
    levels: Union[str, None] = None,
    reverse: Union[str, None] = None,
    overwrite: bool = False,
    ignore: Union[str, None] = None,
    select: Union[str, None] = None,
    threshold: str = "BEST_PRACTICE_SUGGESTION",
    json_file_path: Union[str, None] = None,
    n_jobs: int = 1,
    skip_validate: bool = False,
    detailed: bool = False,
    progress_bar: Union[str, None] = None,
    modules: Union[str, None] = None,
) -> None:
    """
    Run the NWB Inspector via the command line.

    Example Usage
    -------------
    nwbinspector path/to/folder

    nwbinspector path/to/file.nwb

    nwbinspector 000003 --stream

    nwbinspector 000003:sub-YutaMouse39/sub-YutaMouse39_ses-YutaMouse39-150727_behavior+ecephys.nwb  \
      --stream --version-id draft
    """
    path_is_url = path.startswith("https://")
    stream = True if path_is_url else stream

    if stream is True and config is not None and config != "dandi":
        message = (
            "File content originates from DANDI, but a custom config was specified. "
            "The 'dandi' config will be used instead."
        )
        raise ValueError(message)
    elif stream is True and config is None:
        config = "dandi"

    handled_config = config if config is None else load_config(filepath_or_keyword=config)
    handled_levels = ["importance", "file_path"] if levels is None else levels.split(",")
    handled_reverse = [False] * len(handled_levels) if reverse is None else [strtobool(x) for x in reverse.split(",")]
    handled_ignore = ignore if ignore is None else ignore.split(",")
    handled_select = select if select is None else select.split(",")
    handled_importance_threshold = Importance[threshold]
    show_progress_bar = True if progress_bar is None else strtobool(progress_bar)
    handled_modules = [] if modules is None else modules.split(",")

    # Trigger the import of custom checks that have been registered and exposed to their respective modules
    for module in handled_modules:
        importlib.import_module(name=module)

    # Scan entire Dandiset
    if stream and ":" not in path:
        dandiset_id = path
        dandiset_version = version_id

        messages_iterator = inspect_dandiset(
            dandiset_id=dandiset_id,
            dandiset_version=dandiset_version,
            config=handled_config,
            ignore=handled_ignore,
            select=handled_select,
            importance_threshold=handled_importance_threshold,
            skip_validate=skip_validate,
            show_progress_bar=show_progress_bar,
        )
    # Scan a single NWB file in a Dandiset
    elif stream and ":" in path and not path_is_url:
        dandiset_id, dandi_file_path = path.split(":")
        dandiset_version = version_id

        messages_iterator = inspect_dandi_file_path(
            dandi_file_path=dandi_file_path,
            dandiset_id=dandiset_id,
            dandiset_version=dandiset_version,
            config=handled_config,
            ignore=handled_ignore,
            select=handled_select,
            importance_threshold=handled_importance_threshold,
            skip_validate=skip_validate,
        )
    # Scan single NWB file at URL
    elif stream and path_is_url:
        dandi_s3_url = path

        messages_iterator = inspect_url(
            url=dandi_s3_url,
            config=handled_config,
            ignore=handled_ignore,
            select=handled_select,
            importance_threshold=handled_importance_threshold,
            skip_validate=skip_validate,
        )
    # Scan local file/folder
    else:  # stream is False
        messages_iterator = inspect_all(
            path=path,
            config=handled_config,
            ignore=handled_ignore,
            select=handled_select,
            importance_threshold=handled_importance_threshold,
            n_jobs=n_jobs,
            skip_validate=skip_validate,
            progress_bar=show_progress_bar,
        )
    messages = list(messages_iterator)

    if json_file_path is not None:
        if Path(json_file_path).exists() and not overwrite:
            raise FileExistsError(f"The file {json_file_path} already exists! Specify the '-o' flag to overwrite.")
        with open(file=json_file_path, mode="w") as fp:
            json_report = dict(header=_get_report_header(), messages=messages)
            json.dump(obj=json_report, fp=fp, cls=InspectorOutputJSONEncoder)
            print(f"{os.linesep*2}Report saved to {str(Path(json_file_path).absolute())}!{os.linesep}")

    formatted_messages = format_messages(
        messages=messages, levels=handled_levels, reverse=handled_reverse, detailed=detailed
    )
    print_to_console(formatted_messages=formatted_messages)
    if report_file_path is not None:
        save_report(report_file_path=report_file_path, formatted_messages=formatted_messages, overwrite=overwrite)
        print(f"{os.linesep*2}Report saved to {str(Path(report_file_path).absolute())}!{os.linesep}")



"""Primary decorator used on a check function to add it to the registry and automatically parse its output."""

from dataclasses import dataclass
from enum import Enum
from typing import Optional


class Importance(Enum):
    """A definition of the valid importance levels for a given check function."""

    ERROR = 4
    PYNWB_VALIDATION = 3
    CRITICAL = 2
    BEST_PRACTICE_VIOLATION = 1
    BEST_PRACTICE_SUGGESTION = 0


class Severity(Enum):
    """
    A definition of the valid severity levels for the output from a given check function.

    Strictly for internal development that improves report organization; users should never directly see these values.
    """

    HIGH = 2
    LOW = 1


@dataclass
class InspectorMessage:
    """
    The primary output to be returned by every check function.

    Parameters
    ----------
    message : str
        A message that informs the user of the violation.
    severity : Severity, optional
        If a check of non-CRITICAL importance has some basis of comparison, such as magnitude of affected data, then
        the developer of the check may set the severity as Severity.HIGH or Severity.LOW by calling
        `from nwbinspector.register_checks import Severity`. A good example is comparing if h5py.Dataset compression
        has been enabled on smaller vs. larger objects (see nwbinspector/checks/nwb_containers.py for details).

        The user will never directly see this severity, but it will prioritize the order in which check results are
        presented by the NWBInspector.

    importance : Importance
        The Importance level specified by the decorator of the check function.
    check_function_name : str
        The name of the check function the decorator was applied to.
    object_type : str
        The specific class of the instantiated object being inspected.
    object_name : str
        The name of the instantiated object being inspected.
    location : str
        The location relative to the root of the NWBFile where the inspected object may be found.
    file_path : str
        The path of the NWBFile this message pertains to
        Relative to the path called from inspect_nwb, inspect_all, or the path specified at the command line.
    """

    message: str
    importance: Importance = Importance.BEST_PRACTICE_SUGGESTION
    severity: Severity = Severity.LOW
    check_function_name: Optional[str] = None
    object_type: Optional[str] = None
    object_name: Optional[str] = None
    location: Optional[str] = None
    file_path: Optional[str] = None

    def __repr__(self):
        """Representation for InspectorMessage objects according to black format."""
        return "InspectorMessage(\n" + ",\n".join([f"    {k}={v.__repr__()}" for k, v in self.__dict__.items()]) + "\n)"



"""Commonly reused logic for evaluating conditions; must not have external dependencies."""

import json
import os
import re
from functools import lru_cache
from importlib import import_module
from pathlib import Path
from typing import Any, Optional, TypeVar, Union

import h5py
import numpy as np
import zarr
from hdmf.backends.hdf5.h5_utils import H5Dataset
from numpy.typing import ArrayLike
from packaging import version

# TODO: deprecat these in favor of explicit typing
PathType = TypeVar("PathType", str, Path)  # For types that can be either files or folders
FilePathType = TypeVar("FilePathType", str, Path)
OptionalListOfStrings = Optional[list[str]]

dict_regex = r"({.+:.+})"  # TODO: remove this from global scope
MAX_CACHE_ITEMS = 1000  # lru_cache default is 128 calls of matching input/output, but might need more to get use here


@lru_cache(maxsize=MAX_CACHE_ITEMS)
def _cache_data_retrieval_command(
    data: Union[h5py.Dataset, zarr.Array], reduced_selection: tuple[tuple[Optional[int], Optional[int], Optional[int]]]
) -> np.ndarray:
    """LRU caching for _cache_data_selection cannot be applied to list inputs; this expects the tuple or Dataset."""
    selection = tuple([slice(*reduced_slice) for reduced_slice in reduced_selection])  # reconstitute the slices
    return data[selection]


def cache_data_selection(data: Union[h5py.Dataset, ArrayLike], selection: Union[slice, tuple[slice]]) -> np.ndarray:
    """Extract the selection lazily from the data object for efficient caching (most beneficial during streaming)."""
    if isinstance(data, np.memmap):  # np.memmap objects are not hashable - simply return the selection lazily
        return data[selection]
    if not (
        isinstance(data, h5py.Dataset) or isinstance(data, H5Dataset)
    ):  # No need to attempt to cache if data is already in-memory
        # Cast as numpy array for efficient fancy indexing
        # Note that this technically copies the entire data, so could use more than 2x RAM for that object
        return np.array(data)[selection]

    # Slices aren't hashable, but their reduced representation is
    if isinstance(selection, slice):  # A single slice
        reduced_selection = tuple([selection.__reduce__()[1]])
    else:  # Iterable of slices
        reduced_selection = tuple([selection_slice.__reduce__()[1] for selection_slice in selection])
    return _cache_data_retrieval_command(data=data, reduced_selection=reduced_selection)


def format_byte_size(byte_size: int, units: str = "SI") -> str:
    """
    Format a number representing a total number of bytes into a convenient unit.

    Parameters
    ----------
    byte_size : int
        Total number of bytes to format.
    units : str, optional
        Convention for orders of magnitude to apply.
        May be either SI (orders of 1000) or binary (in memory, orders of 1024).
        The default is SI.
    """
    num = float(byte_size)
    prefixes = ["", "K", "M", "G", "T", "P", "E", "Z"]
    if units == "SI":
        order = 1000.0
        suffix = "B"
    elif units == "binary":
        order = 1024.0
        suffix = "iB"
    else:
        raise ValueError("'units' argument must be either 'SI' (for orders of 1000) or 'binary' (for orders of 1024).")
    for prefix in prefixes:
        if abs(num) < order:
            return f"{num:3.2f}{prefix}{suffix}"
        num /= order

    return f"{num:.2f}Y{suffix}"


def is_regular_series(series: np.ndarray, tolerance_decimals: int = 9) -> bool:
    """General purpose function for checking if the difference between all consecutive points in a series are equal."""
    uniq_diff_ts = np.unique(np.diff(series).round(decimals=tolerance_decimals))

    return len(uniq_diff_ts) == 1


def is_ascending_series(series: Union[h5py.Dataset, ArrayLike], nelems: Optional[int] = None) -> bool:
    """General purpose function for determining if a series is monotonic increasing."""
    if isinstance(series, h5py.Dataset):
        data = cache_data_selection(data=series, selection=slice(nelems))
    else:
        data = series[:nelems]

    # Remove NaN values from the series
    data = np.array(data)
    valid_data = data[~np.isnan(data)]

    # Compute the differences between consecutive elements
    differences = np.diff(valid_data)

    return np.all(differences >= 0)


def is_dict_in_string(string: str) -> bool:
    """
    Determine if the string value contains an encoded Python dictionary.

    Can also be the direct results of string casting a dictionary, *e.g.*, ``str(dict(a=1))``.
    """
    return any(re.findall(pattern=dict_regex, string=string))


def is_string_json_loadable(string: str) -> bool:
    """
    Determine if the serialized dictionary is a JSON object.

    Rather than constructing a complicated regex pattern, a simple try/except of the json.load should suffice.
    """
    try:
        json.loads(string)
        return True
    except json.JSONDecodeError:
        return False


def is_module_installed(module_name: str) -> bool:
    """
    Check if the given module is installed on the system.

    Used for lazy imports.
    """
    try:
        import_module(name=module_name)
        return True
    except ModuleNotFoundError:
        return False


def get_package_version(name: str) -> version.Version:
    """
    Retrieve the version of a package regardless of if it has a __version__ attribute set.

    Parameters
    ----------
    name : str
        Name of package.

    Returns
    -------
    version : Version
        The package version as an object from packaging.version.Version, which allows comparison to other versions.
    """
    try:
        from importlib.metadata import version as importlib_version

        package_version = importlib_version(name)
    except ModuleNotFoundError:  # Remove the except clause when minimal supported version becomes 3.8
        from pkg_resources import get_distribution

        package_version = get_distribution(name).version

    return version.parse(package_version)


def calculate_number_of_cpu(requested_cpu: int = 1) -> int:
    """
    Calculate the number CPUs to use with respect to negative slicing and check against maximal available resources.

    Parameters
    ----------
    requested_cpu : int, optional
        The desired number of CPUs to use.

        The default is 1.
    """
    total_cpu = os.cpu_count() or 1  # Annotations say os.cpu_count can return None for some reason
    assert requested_cpu <= total_cpu, f"Requested more CPUs ({requested_cpu}) than are available ({total_cpu})!"
    assert requested_cpu >= -(
        total_cpu - 1
    ), f"Requested fewer CPUs ({requested_cpu}) than are available ({total_cpu})!"
    if requested_cpu > 0:
        return requested_cpu
    else:
        return total_cpu + requested_cpu


def get_data_shape(data: Any, strict_no_data_load: bool = False) -> Optional[tuple[int, ...]]:
    """
    Helper function used to determine the shape of the given array.

    Modified from hdmf.utils.get_data_shape to return shape instead of maxshape.

    In order to determine the shape of nested tuples, lists, and sets, this function
    recursively inspects elements along the dimensions, assuming that the data has a regular,
    rectangular shape. In the case of out-of-core iterators, this means that the first item
    along each dimension would potentially be loaded into memory. Set strict_no_data_load=True
    to enforce that this does not happen, at the cost that we may not be able to determine
    the shape of the array.

    Parameters
    ----------
    data : list, numpy.ndarray, DataChunkIterator, any object that support __len__ or .shape.
        Array for which we should determine the shape.
    strict_no_data_load : bool
        If True and data is an out-of-core iterator, None may be returned.
        If False (default), the first element of data may be loaded into memory.

    Returns
    -------
    data_shape : tuple[int], optional
        Tuple of ints indicating the size of known dimensions.
        Dimensions for which the size is unknown will be set to None.
    """

    def __get_shape_helper(local_data: Any) -> tuple[int, ...]:
        shape = list()
        if hasattr(local_data, "__len__"):
            shape.append(len(local_data))
            if len(local_data):
                el = next(iter(local_data))
                if not isinstance(el, (str, bytes)):
                    shape.extend(__get_shape_helper(el))
        return tuple(shape)

    if hasattr(data, "shape") and data.shape is not None:
        return data.shape
    if isinstance(data, dict):
        return None
    if hasattr(data, "__len__") and not isinstance(data, (str, bytes)):
        if not strict_no_data_load or isinstance(data, (list, tuple, set)):
            return __get_shape_helper(data)

    return None


def strtobool(val: str) -> bool:
    """
    Convert a string representation of truth to True or False.

    True values are 'y', 'yes', 't', 'true', 'on', and '1';
    False values are 'n', 'no', 'f', 'false', 'off', and '0'.
    Raises ValueError if 'val' is anything else.
    """
    if not isinstance(val, str):
        raise TypeError(f"Invalid type of {val!r} - must be str for `strtobool`")
    val = val.lower()
    if val in ("y", "yes", "t", "true", "on", "1"):
        return True
    elif val in ("n", "no", "f", "false", "off", "0"):
        return False
    else:
        raise ValueError(f"Invalid truth value {val!r}")



"""Primary decorator used on a check function to add it to the registry and automatically parse its output."""

from collections.abc import Callable
from functools import wraps
from typing import List, Optional, Union

import h5py
import zarr
from pynwb import NWBFile
from pynwb.ecephys import Device, ElectrodeGroup
from pynwb.file import Subject

from ._types import Importance, InspectorMessage, Severity

available_checks = list()


# TODO: neurodata_type could have annotation hdmf.utils.ExtenderMeta, which seems to apply to all currently checked
# objects. We can wait and see how well that holds up before adding it in officially.
def register_check(importance: Importance, neurodata_type: object) -> Callable:
    """
    Wrap a check function with this decorator to add it to the check registry and automatically parse some output.

    Parameters
    ----------
    importance : Importance
        Importance has three levels:
            CRITICAL
                - potentially incorrect data
            BEST_PRACTICE_VIOLATION
                - very suboptimal data representation
            BEST_PRACTICE_SUGGESTION
                - improvable data representation
    neurodata_type
        The most generic HDMF/PyNWB class the check function applies to.
        Should generally match the type annotation of the check.
        If this check is intended to apply to any general NWBFile object, set neurodata_type to None.
    """

    def register_check_and_auto_parse(check_function: Callable) -> Callable:
        if importance not in [
            Importance.CRITICAL,
            Importance.BEST_PRACTICE_VIOLATION,
            Importance.BEST_PRACTICE_SUGGESTION,
        ]:
            raise ValueError(
                f"Indicated importance ({importance}) of custom check ({check_function.__name__}) is not a valid "
                "importance level! Please choose one of Importance.CRITICAL, Importance.BEST_PRACTICE_VIOLATION, "
                "or Importance.BEST_PRACTICE_SUGGESTION."
            )
        check_function.importance = importance  # type: ignore
        check_function.neurodata_type = neurodata_type  # type: ignore

        @wraps(check_function)
        def auto_parse_some_output(
            *args, **kwargs
        ) -> Union[InspectorMessage, List[Union[InspectorMessage, None]], None]:
            if args:
                neurodata_object = args[0]
            else:
                neurodata_object = kwargs[list(kwargs)[0]]
            output = check_function(*args, **kwargs)

            auto_parsed_result: Union[InspectorMessage, List[Union[InspectorMessage, None]], None] = None
            if isinstance(output, InspectorMessage):
                auto_parsed_result = _auto_parse(
                    check_function=check_function, neurodata_object=neurodata_object, result=output
                )
            elif output is not None:
                auto_parsed_result = list()
                for result in output:
                    auto_parsed_result.append(
                        _auto_parse(check_function=check_function, neurodata_object=neurodata_object, result=result)
                    )
                if not any(auto_parsed_result):
                    auto_parsed_result = None
            return auto_parsed_result

        available_checks.append(auto_parse_some_output)

        return auto_parse_some_output

    return register_check_and_auto_parse


def _auto_parse(
    check_function: Callable, neurodata_object: object, result: Optional[InspectorMessage] = None
) -> Optional[InspectorMessage]:
    """Automatically fill values in the InspectorMessage from the check function."""
    if result is not None:
        auto_parsed_result = result
        if not isinstance(auto_parsed_result.severity, Severity):
            raise ValueError(
                f"Indicated severity ({auto_parsed_result.severity}) of custom check "
                f"({check_function.__name__}) is not a valid severity level! Please choose one of "
                "Severity.HIGH, Severity.LOW, or do not specify any severity."
            )
        auto_parsed_result.importance = check_function.importance  # type: ignore
        auto_parsed_result.check_function_name = check_function.__name__
        auto_parsed_result.object_type = type(neurodata_object).__name__
        auto_parsed_result.object_name = neurodata_object.name  # type: ignore
        auto_parsed_result.location = _parse_location(neurodata_object=neurodata_object)

        return auto_parsed_result

    return None


def _parse_location(neurodata_object: object) -> Optional[str]:
    """Grab the object location from a dataset or a container content that is an dataset object."""
    known_locations = {
        NWBFile: "/",
        Subject: "/general/subject",
        Device: f"/general/devices/{neurodata_object.name}",  # type: ignore
        ElectrodeGroup: f"/general/extracellular_ephys/{neurodata_object.name}",  # type: ignore
        # TODO: add ophys equivalents
    }

    for key, val in known_locations.items():
        if isinstance(neurodata_object, key):
            return val

    # Infer the human-readable path of the object within an NWBFile by tracing its parents
    if neurodata_object.parent is None:  # type: ignore
        return "/"
    # Best solution: object is or has a HDF5 Dataset
    if isinstance(neurodata_object, (h5py.Dataset, zarr.Array)):
        return neurodata_object.name  # type: ignore
    else:
        for field_name, field in neurodata_object.fields.items():  # type: ignore
            if isinstance(field, h5py.Dataset):
                return field.parent.name  # type: ignore
            elif isinstance(field, zarr.Array):
                return field.name.removesuffix(f"/{field_name}")

    return None



import importlib.metadata

from ._registration import available_checks, register_check
from ._types import Importance, Severity, InspectorMessage
from ._configuration import load_config, validate_config, configure_checks
from ._nwb_inspection import (
    inspect_all,
    inspect_nwbfile,
    inspect_nwbfile_object,
    run_checks,
)
from ._formatting import (
    format_messages,
    print_to_console,
    save_report,
    MessageFormatter,
    FormatterOptions,
    InspectorOutputJSONEncoder,
)
from ._organization import organize_messages
from ._dandi_inspection import inspect_dandiset, inspect_dandi_file_path, inspect_url
from .checks import *  # These need to be imported to trigger registration with 'available_checks', but are not exposed

default_check_registry = {check.__name__: check for check in available_checks}

# Still keeping the legacy magic version attribute requested by some users
__version__ = importlib.metadata.version(distribution_name="nwbinspector")

# Note: this is not exposed at this outer level, but is used here to trigger the automatic submodule import
# (otherwise someone would have to import nwbinspector.testing explicitly)
from .testing import check_streaming_tests_enabled  # noqa: F401

__all__ = [
    "available_checks",
    "default_check_registry",
    "register_check",
    "Importance",
    "Severity",
    "InspectorMessage",
    "validate_config",
    "load_config",
    "configure_checks",
    "InspectorOutputJSONEncoder",
    "inspect_dandiset",
    "inspect_dandi_file_path",
    "inspect_url",
    "inspect_all",
    "inspect_nwbfile",
    "inspect_nwbfile_object",
    "run_checks",
    "format_messages",
    "print_to_console",
    "save_report",
    "MessageFormatter",
    "FormatterOptions",
    "organize_messages",
    "__version__",
    # Public submodules
    "checks",
    "testing",
    "utils",
]



from ._dandi import get_s3_urls_and_dandi_paths
from ._nwb import all_of_type, get_nwbfile_path_from_internal_object
from ._read_nwbfile import BACKEND_IO_CLASSES, read_nwbfile, read_nwbfile_and_io

__all__ = [
    "BACKEND_IO_CLASSES",
    "get_s3_urls_and_dandi_paths",
    "all_of_type",
    "get_nwbfile_path_from_internal_object",
    "read_nwbfile",
    "read_nwbfile_and_io",
]



from ._testing import (
    check_hdf5_io_open,
    check_streaming_enabled,
    check_streaming_tests_enabled,
    check_zarr_io_open,
    generate_image_series_testing_files,
    generate_testing_files,
    make_minimal_nwbfile,
)

__all__ = [
    "check_streaming_tests_enabled",
    "check_streaming_enabled",
    "check_hdf5_io_open",
    "check_zarr_io_open",
    "generate_testing_files",
    "generate_image_series_testing_files",
    "make_minimal_nwbfile",
]



"""Check functions that examine general NWBFile metadata."""

import re
from datetime import datetime
from typing import Iterable, Optional

from isodate import Duration, parse_duration
from pynwb import NWBFile, ProcessingModule
from pynwb.file import Subject

from .._registration import Importance, InspectorMessage, register_check
from ..utils import is_module_installed

duration_regex = (
    r"^P(?!$)(\d+(?:\.\d+)?Y)?(\d+(?:\.\d+)?M)?(\d+(?:\.\d+)?W)?(\d+(?:\.\d+)?D)?(T(?=\d)(\d+(?:\.\d+)?H)?(\d+(?:\.\d+)"
    r"?M)?(\d+(?:\.\d+)?S)?)?$"
)
species_form_regex = r"([A-Z][a-z]* [a-z]+)|(http://purl.obolibrary.org/obo/NCBITaxon_\d+)"

PROCESSING_MODULE_CONFIG = ["ophys", "ecephys", "icephys", "behavior", "misc", "ogen", "retinotopy"]


@register_check(importance=Importance.BEST_PRACTICE_SUGGESTION, neurodata_type=NWBFile)
def check_session_start_time_old_date(nwbfile: NWBFile) -> Optional[InspectorMessage]:
    """
    Check if the session_start_time was set to an appropriate value.

    Best Practice: :ref:`best_practice_global_time_reference`
    """
    session_start_time = nwbfile.session_start_time
    dummy_time = datetime(1980, 1, 1)

    tzinfo = session_start_time.tzinfo
    if tzinfo is not None:
        dummy_time = dummy_time.astimezone(tzinfo)

    if session_start_time <= dummy_time:
        return InspectorMessage(
            message=(f"The session_start_time ({session_start_time}) may not be set to the true date of the recording.")
        )

    return None


@register_check(importance=Importance.CRITICAL, neurodata_type=NWBFile)
def check_session_start_time_future_date(nwbfile: NWBFile) -> Optional[InspectorMessage]:
    """
    Check if the session_start_time was set to an appropriate value.

    Best Practice: :ref:`best_practice_global_time_reference`
    """
    session_start_time = nwbfile.session_start_time
    current_time = datetime.now()
    tzinfo = session_start_time.tzinfo
    if session_start_time.tzinfo is not None:
        current_time = current_time.astimezone(tzinfo)

    if session_start_time >= current_time:
        return InspectorMessage(
            message=f"The session_start_time ({session_start_time}) is set to a future date and time."
        )

    return None


@register_check(importance=Importance.BEST_PRACTICE_SUGGESTION, neurodata_type=NWBFile)
def check_experimenter_exists(nwbfile: NWBFile) -> Optional[InspectorMessage]:
    """
    Check if an experimenter has been added for the session.

    Best Practice: :ref:`best_practice_experimenter`
    """
    if not nwbfile.experimenter:
        return InspectorMessage(message="Experimenter is missing.")

    return None


@register_check(importance=Importance.BEST_PRACTICE_SUGGESTION, neurodata_type=NWBFile)
def check_experimenter_form(nwbfile: NWBFile) -> Optional[Iterable[InspectorMessage]]:
    """
    Check the text form of each experimenter to see if it matches the DANDI regex pattern.

    Best Practice: :ref:`best_practice_experimenter`
    """
    if nwbfile.experimenter is None:
        return None
    if is_module_installed(module_name="dandi"):
        from dandischema.models import (
            NAME_PATTERN,  # for most up to date version of the regex
        )
    else:
        NAME_PATTERN = r"^([\w\s\-\.']+),\s+([\w\s\-\.']+)$"  # copied on 7/12/22

    for experimenter in nwbfile.experimenter:
        experimenter = experimenter.decode() if isinstance(experimenter, bytes) else experimenter
        if re.match(string=experimenter, pattern=NAME_PATTERN) is None:
            yield InspectorMessage(
                message=(
                    f"The name of experimenter '{experimenter}' does not match any of the accepted DANDI forms: "
                    "'LastName, Firstname', 'LastName, FirstName MiddleInitial.' or 'LastName, FirstName, MiddleName'."
                )
            )

    return None


@register_check(importance=Importance.BEST_PRACTICE_SUGGESTION, neurodata_type=NWBFile)
def check_experiment_description(nwbfile: NWBFile) -> Optional[InspectorMessage]:
    """Check if a description has been added for the session."""
    if not nwbfile.experiment_description:
        return InspectorMessage(message="Experiment description is missing.")

    return None


@register_check(importance=Importance.BEST_PRACTICE_SUGGESTION, neurodata_type=NWBFile)
def check_institution(nwbfile: NWBFile) -> Optional[InspectorMessage]:
    """Check if a description has been added for the session."""
    if not nwbfile.institution:
        return InspectorMessage(message="Metadata /general/institution is missing.")

    return None


@register_check(importance=Importance.BEST_PRACTICE_SUGGESTION, neurodata_type=NWBFile)
def check_keywords(nwbfile: NWBFile) -> Optional[InspectorMessage]:
    """Check if keywords have been added for the session."""
    if not nwbfile.keywords:
        return InspectorMessage(message="Metadata /general/keywords is missing.")

    return None


@register_check(importance=Importance.BEST_PRACTICE_SUGGESTION, neurodata_type=NWBFile)
def check_subject_exists(nwbfile: NWBFile) -> Optional[InspectorMessage]:
    """Check if subject exists."""
    if nwbfile.subject is None:
        return InspectorMessage(message="Subject is missing.")

    return None


@register_check(importance=Importance.BEST_PRACTICE_SUGGESTION, neurodata_type=NWBFile)
def check_doi_publications(nwbfile: NWBFile) -> Optional[Iterable[InspectorMessage]]:
    """Check if related_publications has been properly added as 'doi: ###' or an external 'doi' link."""
    valid_starts = ["doi:", "http://dx.doi.org/", "https://doi.org/"]

    if not nwbfile.related_publications:
        return None
    for publication in nwbfile.related_publications:
        publication = publication.decode() if isinstance(publication, bytes) else publication
        if not any((publication.startswith(valid_start) for valid_start in valid_starts)):
            yield InspectorMessage(
                message=(
                    f"Metadata /general/related_publications '{publication}' does not start with 'doi: ###' and is "
                    "not an external 'doi' link."
                )
            )

    return None


@register_check(importance=Importance.BEST_PRACTICE_SUGGESTION, neurodata_type=Subject)
def check_subject_age(subject: Subject) -> Optional[InspectorMessage]:
    """Check if the Subject age is in ISO 8601 or our extension of it for ranges."""
    if subject.age is None and subject.date_of_birth is None:
        return InspectorMessage(
            message="Subject is missing age and date_of_birth. Please specify at least one of these fields."
        )
    elif subject.age is None and subject.date_of_birth is not None:
        return None
    if re.fullmatch(pattern=duration_regex, string=subject.age):
        return None

    if "/" in subject.age:
        subject_lower_age_bound, subject_upper_age_bound = subject.age.split("/")

        if re.fullmatch(pattern=duration_regex, string=subject_lower_age_bound) and (
            re.fullmatch(pattern=duration_regex, string=subject_upper_age_bound) or subject_upper_age_bound == ""
        ):
            return None

    return InspectorMessage(
        message=(
            f"Subject age, '{subject.age}', does not follow ISO 8601 duration format, e.g. 'P2Y' for 2 years "
            "or 'P23W' for 23 weeks. You may also specify a range using a '/' separator, e.g., 'P1D/P3D' for an "
            "age range somewhere from 1 to 3 days. If you cannot specify the upper bound of the range, "
            "you may leave the right side blank, e.g., 'P90Y/' means 90 years old or older."
        )
    )


@register_check(importance=Importance.BEST_PRACTICE_SUGGESTION, neurodata_type=Subject)
def check_subject_proper_age_range(subject: Subject) -> Optional[InspectorMessage]:
    """
    Check if the Subject age, if specified as duration range (e.g., 'P1D/P3D'), has properly increasing bounds.

    Best Practice: :ref:`best_practice_subject_age`
    """
    if subject.age is not None and "/" in subject.age:
        subject_lower_age_bound, subject_upper_age_bound = subject.age.split("/")

        if re.fullmatch(pattern=duration_regex, string=subject_lower_age_bound) and re.fullmatch(
            pattern=duration_regex, string=subject_upper_age_bound
        ):
            lower = parse_duration(subject_lower_age_bound)
            if isinstance(lower, Duration):
                lower = lower.totimedelta(end=datetime.now())

            upper = parse_duration(subject_upper_age_bound)
            if isinstance(upper, Duration):
                upper = upper.totimedelta(end=datetime.now())

            if lower >= upper:
                return InspectorMessage(
                    message=(
                        f"The durations of the Subject age range, '{subject.age}', are not strictly increasing. "
                        "The upper (right) bound should be a longer duration than the lower (left) bound."
                    )
                )

    return None


@register_check(importance=Importance.BEST_PRACTICE_SUGGESTION, neurodata_type=Subject)
def check_subject_id_exists(subject: Subject) -> Optional[InspectorMessage]:
    """
    Check if subject_id is defined.

    Best Practice: :ref:`best_practice_subject_id`
    """
    if subject.subject_id is None:
        return InspectorMessage(message="subject_id is missing.")

    return None


def _check_subject_sex_defaults(sex: str) -> Optional[InspectorMessage]:
    """Check if the subject sex has been specified properly for the C. elegans species."""
    if sex not in ("M", "F", "O", "U"):
        return InspectorMessage(
            message="Subject.sex should be one of: 'M' (male), 'F' (female), 'O' (other), or 'U' (unknown)."
        )

    return None


def _check_subject_sex_c_elegans(sex: str) -> Optional[InspectorMessage]:
    """Check if the subject sex has been specified properly for the C. elegans species."""
    if sex not in ("XO", "XX"):
        return InspectorMessage(message="For C. elegans, Subject.sex should be 'XO' (male) or 'XX' (hermaphrodite).")

    return None


@register_check(importance=Importance.BEST_PRACTICE_SUGGESTION, neurodata_type=Subject)
def check_subject_sex(subject: Subject) -> Optional[InspectorMessage]:
    """
    Check if the subject sex has been specified and ensure that it has has the correct form depending on the species.

    Best Practice: :ref:`best_practice_subject_sex`
    """
    if subject and not subject.sex:
        return InspectorMessage(message="Subject.sex is missing.")
    if subject.species in ("Caenorhabditis elegans", "C. elegans"):
        return _check_subject_sex_c_elegans(sex=subject.sex)
    else:
        return _check_subject_sex_defaults(sex=subject.sex)

    return None


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=Subject)
def check_subject_species_exists(subject: Subject) -> Optional[InspectorMessage]:
    """
    Check if the subject species has been specified.

    Best Practice: :ref:`best_practice_subject_species`
    """
    if not subject.species:
        return InspectorMessage(message="Subject species is missing.")

    return None


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=Subject)
def check_subject_species_form(subject: Subject) -> Optional[InspectorMessage]:
    """
    Check if the subject species follows latin binomial form or is a link to an NCBI taxonomy in the form of a Term IRI.

    The Term IRI can be found at the https://ontobee.org/ database.

    Best Practice: :ref:`best_practice_subject_species`
    """
    if subject.species and not re.fullmatch(species_form_regex, subject.species):
        return InspectorMessage(
            message=(
                f"Subject species '{subject.species}' should either be in Latin binomial form (e.g., 'Mus musculus' and "
                "'Homo sapiens') or be a NCBI taxonomy link (e.g., 'http://purl.obolibrary.org/obo/NCBITaxon_280675')."
            ),
        )

    return None


@register_check(importance=Importance.BEST_PRACTICE_SUGGESTION, neurodata_type=ProcessingModule)
def check_processing_module_name(processing_module: ProcessingModule) -> Optional[InspectorMessage]:
    """Check if the name of a processing module is of a valid modality."""
    if processing_module.name not in PROCESSING_MODULE_CONFIG:
        return InspectorMessage(
            message=(
                f"Processing module is named {processing_module.name}. It is recommended to use the "
                f"schema module names: {', '.join(PROCESSING_MODULE_CONFIG)}"
            )
        )

    return None


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=NWBFile)
def check_session_id_no_slashes(nwbfile: NWBFile) -> Optional[InspectorMessage]:
    """
    Check if session_id contains slash characters, which can cause problems when constructing paths in DANDI.

    Best Practice: :ref:`best_practice_session_id`
    """
    if nwbfile.session_id and "/" in nwbfile.session_id:
        return InspectorMessage(
            message=(
                f"The session_id '{nwbfile.session_id}' contains slash character(s) '/', which can cause problems "
                f"when constructing paths in DANDI. Please replace slashes with another character (e.g., '-' or '_')."
            )
        )

    return None


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=Subject)
def check_subject_id_no_slashes(subject: Subject) -> Optional[InspectorMessage]:
    """
    Check if subject_id contains slash characters, which can cause problems when constructing paths in DANDI.

    Best Practice: :ref:`best_practice_subject_id`
    """
    if subject.subject_id and "/" in subject.subject_id:
        return InspectorMessage(
            message=(
                f"The subject_id '{subject.subject_id}' contains slash character(s) '/', which can cause problems "
                f"when constructing paths in DANDI. Please replace slashes with another character (e.g., '-' or '_')."
            )
        )

    return None



from typing import Optional

from pynwb.ogen import OptogeneticSeries, OptogeneticStimulusSite

from .._registration import Importance, InspectorMessage, register_check


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=OptogeneticStimulusSite)
def check_optogenetic_stimulus_site_has_optogenetic_series(
    ogen_site: OptogeneticStimulusSite,
) -> Optional[InspectorMessage]:
    """Check if an optogenetic stimulus site has an optogenetic series linked to it."""
    nwbfile = ogen_site.get_ancestor("NWBFile")
    for obj in nwbfile.objects.values():
        if isinstance(obj, OptogeneticSeries):
            if obj.site == ogen_site:
                return None

    return InspectorMessage(message="OptogeneticStimulusSite is not referenced by any OptogeneticSeries.")



"""Check functions that can apply to any descendant of TimeSeries."""

from typing import Optional

import numpy as np
from pynwb import TimeSeries
from pynwb.ecephys import SpikeEventSeries
from pynwb.image import ImageSeries, IndexSeries

from .._registration import Importance, InspectorMessage, Severity, register_check
from ..utils import get_data_shape, is_ascending_series, is_regular_series


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=TimeSeries)
def check_regular_timestamps(
    time_series: TimeSeries, time_tolerance_decimals: int = 9, gb_severity_threshold: float = 1.0
) -> Optional[InspectorMessage]:
    """If the TimeSeries uses timestamps, check if they are regular (i.e., they have a constant rate)."""
    if (
        time_series.timestamps is not None
        and len(time_series.timestamps) > 2
        and is_regular_series(series=time_series.timestamps, tolerance_decimals=time_tolerance_decimals)
        and (time_series.timestamps[1] - time_series.timestamps[0]) != 0
    ):
        timestamps = np.array(time_series.timestamps)
        if timestamps.size * timestamps.dtype.itemsize > gb_severity_threshold * 1e9:
            severity = Severity.HIGH
        else:
            severity = Severity.LOW
        return InspectorMessage(
            severity=severity,
            message=(
                "TimeSeries appears to have a constant sampling rate. "
                f"Consider specifying starting_time={time_series.timestamps[0]} "
                f"and rate={1 / (time_series.timestamps[1] - time_series.timestamps[0])} instead of timestamps."
            ),
        )

    return None


@register_check(importance=Importance.CRITICAL, neurodata_type=TimeSeries)
def check_data_orientation(time_series: TimeSeries) -> Optional[InspectorMessage]:
    """If the TimeSeries has data, check if the longest axis (almost always time) is also the zero-axis."""

    # Skip this check for SpikeEventSeries since its data structure is (events, channels, waveform samples)
    # and it's valid for the number of waveform samples to be larger than the number of events
    if isinstance(time_series, SpikeEventSeries):
        return None

    if time_series.data is None:
        return None

    data_shape = get_data_shape(time_series.data)
    if data_shape is None:
        return None

    if any(np.array(data_shape[1:]) > data_shape[0]):
        return InspectorMessage(
            message=(
                "Data may be in the wrong orientation. Time should be in the first dimension, and is usually the "
                "longest dimension. Here, another dimension is longer."
            ),
        )

    return None


@register_check(importance=Importance.CRITICAL, neurodata_type=TimeSeries)
def check_timestamps_match_first_dimension(time_series: TimeSeries) -> Optional[InspectorMessage]:
    """
    If the TimeSeries has timestamps, check if their length is the same as the zero-axis of data.

    Best Practice: :ref:`best_practice_data_orientation`
    """
    if time_series.data is None or time_series.timestamps is None:
        return None

    data_shape = get_data_shape(time_series.data)
    if data_shape is None:
        return None

    timestamps_shape = get_data_shape(time_series.timestamps)
    if timestamps_shape is None:
        return None

    if getattr(time_series, "external_file", None) is not None and data_shape[0] == 0:
        return None

    # A very specific edge case where this has been allowed, though much more preferable
    # to use a stack of Images rather than an ImageSeries
    if (
        isinstance(time_series, ImageSeries)
        and len(time_series.timestamps) == 0
        and time_series.get_ancestor("NWBFile") is not None
    ):
        for neurodata_object in time_series.get_ancestor("NWBFile").objects.values():
            if isinstance(neurodata_object, IndexSeries) and neurodata_object.indexed_timeseries == time_series:
                return None

    if data_shape[0] != timestamps_shape[0]:
        return InspectorMessage(
            message=(
                f"The length of the first dimension of data ({data_shape[0]}) "
                f"does not match the length of timestamps ({timestamps_shape[0]})."
            )
        )

    return None


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=TimeSeries)
def check_timestamps_ascending(time_series: TimeSeries, nelems: Optional[int] = 200) -> Optional[InspectorMessage]:
    """Check that the values in the timestamps array are strictly increasing."""
    if time_series.timestamps is not None and not is_ascending_series(time_series.timestamps, nelems=nelems):
        return InspectorMessage(f"{time_series.name} timestamps are not ascending.")

    return None


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=TimeSeries)
def check_timestamps_without_nans(time_series: TimeSeries, nelems: Optional[int] = 200) -> Optional[InspectorMessage]:
    """Check if there are NaN values in the timestamps array."""
    if time_series.timestamps is not None and np.isnan(time_series.timestamps[:nelems]).any():
        return InspectorMessage(message=f"{time_series.name} timestamps contain NaN values.")

    return None


@register_check(importance=Importance.BEST_PRACTICE_SUGGESTION, neurodata_type=TimeSeries)
def check_timestamp_of_the_first_sample_is_not_negative(time_series: TimeSeries) -> Optional[InspectorMessage]:
    """
    Check that the timestamp of the first sample is not negative.

    Best Practice: :ref:`best_practice_avoid_negative_timestamps`
    """

    first_timestamp = time_series.starting_time if time_series.starting_time is not None else time_series.timestamps[0]
    if first_timestamp < 0:
        message = (
            "Timestamps should not be negative. This usually indicates a temporal misalignment of the data. "
            "It is recommended to align the `session_start_time` or `timestamps_reference_time` to be the earliest time value that occurs in the data, and shift all other signals accordingly."
        )
        return InspectorMessage(message=message)

    return None


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=TimeSeries)
def check_missing_unit(time_series: TimeSeries) -> Optional[InspectorMessage]:
    """
    Check if the TimeSeries.unit field is empty.

    Best Practice: :ref:`best_practice_unit_of_measurement`
    """
    if not time_series.unit:
        return InspectorMessage(
            message="Missing text for attribute 'unit'. Please specify the scientific unit of the 'data'."
        )

    return None


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=TimeSeries)
def check_resolution(time_series: TimeSeries) -> Optional[InspectorMessage]:
    """Check the resolution value of a TimeSeries for proper format (-1.0 or NaN for unknown)."""
    if time_series.resolution is None or time_series.resolution == -1.0:
        return None
    if time_series.resolution <= 0:
        return InspectorMessage(
            message=f"'resolution' should use -1.0 or NaN for unknown instead of {time_series.resolution}."
        )

    return None


@register_check(importance=Importance.CRITICAL, neurodata_type=TimeSeries)
def check_rate_is_not_zero(time_series: TimeSeries) -> Optional[InspectorMessage]:
    if time_series.data is None:
        return None

    data_shape = get_data_shape(time_series.data)
    if data_shape is None:
        return None

    if time_series.rate == 0.0 and data_shape[0] > 1:
        return InspectorMessage(
            f"{time_series.name} has a sampling rate value of 0.0Hz but the series has more than one frame."
        )

    return None


@register_check(importance=Importance.CRITICAL, neurodata_type=TimeSeries)
def check_rate_is_positive(time_series: TimeSeries) -> Optional[InspectorMessage]:
    if not hasattr(time_series, "rate"):
        return None

    if time_series.rate is not None and time_series.rate < 0.0:
        return InspectorMessage(
            message=f"{time_series.name} has a negative sampling rate value of {time_series.rate}Hz which is not valid."
        )

    return None



import importlib.metadata

import warnings

message = (
    "The 'version' submodule has been deprecated. "
    "Please import the version using `importlib.metadata.version('nwbinspector')`."
)

warnings.warn(message=message, category=DeprecationWarning, stacklevel=2)

# Still keep imports functional with warning for soft deprecation cycle
# TODO: remove after 9/15/2024
__version__ = importlib.metadata.version(distribution_name="nwbinspector")



"""Check functions specific to optical electrophysiology neurodata types."""

from typing import Iterable, Optional

from pynwb.ophys import (
    ImagingPlane,
    OpticalChannel,
    PlaneSegmentation,
    RoiResponseSeries,
)

from .._registration import Importance, InspectorMessage, register_check
from ..utils import get_data_shape

MIN_LAMBDA = 10.0  # trigger warnings for wavelength values less than this value


@register_check(importance=Importance.CRITICAL, neurodata_type=RoiResponseSeries)
def check_roi_response_series_dims(roi_response_series: RoiResponseSeries) -> Optional[InspectorMessage]:
    """
    Check the dimensions of an ROI series to ensure the time axis is the correct dimension.

    Best Practice: :ref:`best_practice_data_orientation`
    """
    data = roi_response_series.data
    rois = roi_response_series.rois

    data_shape = get_data_shape(data, strict_no_data_load=True)

    if data_shape and len(data_shape) == 2 and data_shape[1] != len(rois.data):
        if data_shape[0] == len(rois.data):
            return InspectorMessage(
                message=(
                    "The second dimension of data does not match the length of rois, "
                    "but instead the first does. Data is oriented incorrectly and should be transposed."
                )
            )
        return InspectorMessage(
            message="The second dimension of data does not match the length of rois. Your data may be transposed."
        )

    return None


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=RoiResponseSeries)
def check_roi_response_series_link_to_plane_segmentation(
    roi_response_series: RoiResponseSeries,
) -> Optional[InspectorMessage]:
    """
    Check that each ROI response series links to a plane segmentation.

    Best Practice: TODO
    """
    if not isinstance(roi_response_series.rois.table, PlaneSegmentation):
        return InspectorMessage(message="rois field does not point to a PlaneSegmentation table.")

    return None


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=OpticalChannel)
def check_emission_lambda_in_nm(optical_channel: OpticalChannel) -> Optional[InspectorMessage]:
    """
    Check that emission lambda is in feasible range for unit nanometers.

    Best Practice: :ref:`best_practice_unit_of_measurement`
    """
    if optical_channel.emission_lambda < MIN_LAMBDA:
        return InspectorMessage(f"emission lambda of {optical_channel.emission_lambda} should be in units of nm.")

    return None


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=ImagingPlane)
def check_excitation_lambda_in_nm(imaging_plane: ImagingPlane) -> Optional[InspectorMessage]:
    """
    Check that emission lambda is in feasible range for unit nanometers.

    Best Practice: :ref:`best_practice_unit_of_measurement`
    """
    if imaging_plane.excitation_lambda < MIN_LAMBDA:
        return InspectorMessage(f"excitation lambda of {imaging_plane.excitation_lambda} should be in units of nm.")

    return None


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=PlaneSegmentation)
def check_plane_segmentation_image_mask_shape_against_ref_images(
    plane_segmentation: PlaneSegmentation,
) -> Optional[Iterable[InspectorMessage]]:
    """Check that image masks and reference images have the same shape."""
    if plane_segmentation.reference_images and "image_mask" in plane_segmentation.colnames:
        mask_shape = plane_segmentation["image_mask"].shape[1:]
        for ref_image in plane_segmentation.reference_images:
            if mask_shape != ref_image.data.shape[1:]:
                yield InspectorMessage(
                    f"image_mask of shape {mask_shape} does not match reference image {ref_image.name} with shape"
                    f" {ref_image.data.shape[1:]}."
                )

    return None



"""Helper functions related to DANDI for internal use that rely on external dependencies (i.e., dandi)."""

import re
from concurrent.futures import ProcessPoolExecutor, as_completed
from typing import Optional

from ..utils import calculate_number_of_cpu, is_module_installed


def get_s3_urls_and_dandi_paths(dandiset_id: str, version_id: Optional[str] = None, n_jobs: int = 1) -> dict[str, str]:
    """
    Collect S3 URLS from a DANDISet ID.

    Returns dictionary that maps each S3 url to the displayed file path on the DANDI archive content page.
    """
    assert is_module_installed(module_name="dandi"), "You must install DANDI to get S3 paths (pip install dandi)."
    from dandi.dandiapi import DandiAPIClient

    assert re.fullmatch(
        pattern="^[0-9]{6}$", string=dandiset_id
    ), "The specified 'path' is not a proper DANDISet ID. It should be a six-digit numeric identifier."

    s3_urls_to_dandi_paths = dict()
    n_jobs = calculate_number_of_cpu(requested_cpu=n_jobs)
    if n_jobs != 1:
        with DandiAPIClient() as client:
            dandiset = client.get_dandiset(dandiset_id=dandiset_id, version_id=version_id)
            max_workers = n_jobs if n_jobs > 0 else None
            with ProcessPoolExecutor(max_workers=max_workers) as executor:
                futures = []
                for asset in dandiset.get_assets():
                    if asset.path.split(".")[-1] == "nwb":
                        futures.append(
                            executor.submit(
                                _get_content_url_and_path, asset=asset, follow_redirects=1, strip_query=True
                            )
                        )
                    for future in as_completed(futures):
                        s3_urls_to_dandi_paths.update(future.result())
    else:
        with DandiAPIClient() as client:
            dandiset = client.get_dandiset(dandiset_id=dandiset_id, version_id=version_id)
            for asset in dandiset.get_assets():
                if asset.path.split(".")[-1] == "nwb":
                    s3_urls_to_dandi_paths.update(_get_content_url_and_path(asset=asset))
    return s3_urls_to_dandi_paths


def _get_content_url_and_path(
    asset: "dandi.dandiapi.BaseRemoteAsset",  # type: ignore
    follow_redirects: int = 1,
    strip_query: bool = True,
) -> dict[str, str]:
    """
    Private helper function for parallelization in 'get_s3_urls_and_dandi_paths'.

    Must be globally defined (not as a part of get_s3_urls..) in order to be pickled.
    """
    return {asset.get_content_url(follow_redirects=1, strip_query=True): asset.path}



"""Check functions specific to extracellular electrophysiology neurodata types."""

from typing import Optional

import numpy as np
from pynwb.ecephys import ElectricalSeries, SpikeEventSeries
from pynwb.misc import Units

from .._registration import Importance, InspectorMessage, register_check
from ..utils import get_data_shape

NELEMS = 200


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=Units)
def check_negative_spike_times(units_table: Units) -> Optional[InspectorMessage]:
    """Check if the Units table contains negative spike times."""
    if "spike_times" not in units_table:
        return None
    if np.any(np.asarray(units_table["spike_times"].target.data[:]) < 0):
        return InspectorMessage(
            message=(
                "This Units table contains negative spike times. Time should generally be aligned to the earliest "
                "time reference in the NWBFile."
            )
        )

    return None


@register_check(importance=Importance.CRITICAL, neurodata_type=ElectricalSeries)
def check_electrical_series_dims(electrical_series: ElectricalSeries) -> Optional[InspectorMessage]:
    """
    Use the length of the linked electrode region to check the data orientation.

    Best Practice: :ref:`best_practice_data_orientation`
    """
    data = electrical_series.data
    electrodes = electrical_series.electrodes

    data_shape = get_data_shape(data, strict_no_data_load=True)

    # For SpikeEventSeries, only perform the check for 3D data
    if isinstance(electrical_series, SpikeEventSeries):
        if data_shape and len(data_shape) == 3 and data_shape[1] != len(electrodes.data):
            if data_shape[0] == len(electrodes.data):
                return InspectorMessage(
                    message=(
                        "The second dimension of data does not match the length of electrodes, "
                        "but instead the first does. Data is oriented incorrectly and should be transposed."
                    )
                )
            return InspectorMessage(
                message=(
                    "The second dimension of data does not match the length of electrodes. Your data may be transposed."
                )
            )
        # Do not warn for 2D SpikeEventSeries
        return None

    # For other ElectricalSeries, keep the original logic
    if data_shape and len(data_shape) == 2 and data_shape[1] != len(electrodes.data):
        if data_shape[0] == len(electrodes.data):
            return InspectorMessage(
                message=(
                    "The second dimension of data does not match the length of electrodes, "
                    "but instead the first does. Data is oriented incorrectly and should be transposed."
                )
            )
        return InspectorMessage(
            message=(
                "The second dimension of data does not match the length of electrodes. Your data may be transposed."
            )
        )

    return None


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=ElectricalSeries)
def check_electrical_series_reference_electrodes_table(
    electrical_series: ElectricalSeries,
) -> Optional[InspectorMessage]:
    """
    Check that the 'electrodes' of an ElectricalSeries references the ElectrodesTable.

    Best Practice: TODO
    """
    if electrical_series.electrodes.table.name != "electrodes":
        return InspectorMessage(message="electrodes does not  reference an electrodes table.")

    return None


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=Units)
def check_spike_times_not_in_unobserved_interval(units_table: Units, nunits: int = 4) -> Optional[InspectorMessage]:
    """Check if a Units table has spike times that occur outside of observed intervals."""
    if not units_table.obs_intervals:
        return None
    for unit_spike_times, unit_obs_intervals in zip(
        units_table["spike_times"][:nunits], units_table["obs_intervals"][:nunits]
    ):
        spike_times_array = np.array(unit_spike_times)
        if not all(
            sum(
                [
                    np.logical_and(start <= spike_times_array, spike_times_array <= stop)
                    for start, stop in unit_obs_intervals
                ]
            )
        ):
            return InspectorMessage(
                message=(
                    "This Units table contains spike times that occur during periods of time not labeled as being "
                    "observed intervals."
                )
            )

    return None


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=Units)
def check_ascending_spike_times(units_table: Units, nelems: Optional[int] = NELEMS) -> Optional[InspectorMessage]:
    """
    Check that the values in the timestamps array are strictly increasing.

    Best Practice :ref:`best_practice_ascending_spike_times`
    """
    if "spike_times" not in units_table:
        return None

    for unit_id in range(len(units_table)):
        spike_times = units_table["spike_times"][unit_id]
        if nelems is not None:
            spike_times = spike_times[:nelems]
        if not np.all(np.diff(spike_times) >= 0):
            return InspectorMessage(
                message=(
                    f"Unit {unit_id} contains non-ascending spike times. "
                    "Spike times should be sorted in ascending order."
                )
            )
    return None



"""Helper functions related to NWB for internal use that rely on external dependencies (i.e., pynwb)."""

from typing import Iterable, Type

from pynwb import NWBFile


def all_of_type(nwbfile: NWBFile, neurodata_type: Type) -> Iterable[object]:
    """Iterate over all objects inside an NWBFile object and return those that match the given neurodata_type."""
    for neurodata_object in nwbfile.objects.values():
        if isinstance(neurodata_object, neurodata_type):
            yield neurodata_object


def get_nwbfile_path_from_internal_object(neurodata_object: object) -> str:
    """Determine the file path on disk for a NWBFile given only an internal object of that file."""
    if isinstance(neurodata_object, NWBFile):
        return neurodata_object.container_source

    return neurodata_object.get_ancestor("NWBFile").container_source  # type: ignore



import warnings

message = (
    "The submodule 'nwbinspector.tools.dandi' has been deprecated. "
    "Please perform any imports directly from 'nwbinspector.tools' instead."
)

warnings.warn(message=message, category=DeprecationWarning, stacklevel=2)

# Still keep imports functional with warning for soft deprecation cycle
# TODO: remove after 9/15/2024
from .._dandi import *



import pathlib
from typing import Iterable, Literal, Union
from warnings import filterwarnings

import h5py
import pynwb

from ._configuration import load_config, validate_config
from ._nwb_inspection import inspect_nwbfile_object
from ._types import Importance, InspectorMessage


def inspect_dandiset(
    *,
    dandiset_id: str,
    dandiset_version: Union[str, Literal["draft"], None] = None,
    config: Union[str, pathlib.Path, dict, Literal["dandi"], None] = None,
    checks: Union[list, None] = None,
    ignore: Union[list[str], None] = None,
    select: Union[list[str], None] = None,
    importance_threshold: Union[str, Importance] = Importance.BEST_PRACTICE_SUGGESTION,
    skip_validate: bool = False,
    show_progress_bar: bool = True,
    client: Union["dandi.dandiapi.DandiAPIClient", None] = None,  # type: ignore
) -> Iterable[Union[InspectorMessage, None]]:
    """
    Inspect a Dandiset for common issues.

    Parameters
    ----------
    dandiset_id : six-digit string, "draft", or None
        The six-digit ID of the Dandiset to inspect.
    dandiset_version : string
        The specific published version of the Dandiset to inspect.
        If None, the latest version is used.
        If there are no published versions, then 'draft' is used instead.
    config : file path, dictionary, or "dandi", default: "dandi"
        If a file path, loads the dictionary configuration from the file.
        If a dictionary, it must be valid against the configuration schema.
        If "dandi", uses the requirements for DANDI validation.
    checks : list, optional
        list of checks to run
    ignore: list, optional
        Names of functions to skip.
    select: list, optional
        Names of functions to pick out of available checks.
    importance_threshold : string or Importance, optional
        Ignores tests with an assigned importance below this threshold.
        Importance has three levels:

            CRITICAL
                - potentially incorrect data
            BEST_PRACTICE_VIOLATION
                - very suboptimal data representation
            BEST_PRACTICE_SUGGESTION
                - improvable data representation

        The default is the lowest level, BEST_PRACTICE_SUGGESTION.
    skip_validate : bool, default: False
        Skip the PyNWB validation step.
        This may be desired for older NWBFiles (< schema version v2.10).
    show_progress_bar : bool, optional
        Whether to display a progress bar while scanning the assets on the Dandiset.
    client: dandi.dandiapi.DandiAPIClient
        The client object can be passed to avoid re-instantiation over an iteration.
    """
    config = config or "dandi"

    if client is None:
        import dandi.dandiapi

        client = dandi.dandiapi.DandiAPIClient()

    dandiset = client.get_dandiset(dandiset_id=dandiset_id, version_id=dandiset_version)

    nwb_assets = [asset for asset in dandiset.get_assets() if ".nwb" in pathlib.Path(asset.path).suffixes]

    nwb_assets_iterator = nwb_assets
    if show_progress_bar:
        import tqdm

        nwb_assets_iterator = tqdm.tqdm(
            iterable=nwb_assets, total=len(nwb_assets), desc="Inspecting NWB files", unit="file", position=0, leave=True
        )

    for asset in nwb_assets_iterator:
        asset_url = asset.get_content_url(follow_redirects=1, strip_query=True)

        for message in inspect_url(
            url=asset_url,
            config=config,
            checks=checks,
            ignore=ignore,
            select=select,
            importance_threshold=importance_threshold,
            skip_validate=skip_validate,
        ):
            message.file_path = asset.path  # type: ignore
            yield message


def inspect_dandi_file_path(
    *,
    dandi_file_path: str,
    dandiset_id: str,
    dandiset_version: Union[str, Literal["draft"], None] = None,
    config: Union[str, pathlib.Path, dict, Literal["dandi"], None] = "dandi",
    checks: Union[list, None] = None,
    ignore: Union[list[str], None] = None,
    select: Union[list[str], None] = None,
    importance_threshold: Union[str, Importance] = Importance.BEST_PRACTICE_SUGGESTION,
    skip_validate: bool = False,
    client: Union["dandi.dandiapi.DandiAPIClient", None] = None,  # type: ignore
) -> Iterable[Union[InspectorMessage, None]]:
    """
    Inspect a Dandifile for common issues.

    Parameters
    ----------
    dandi_file_path : string
        The path to the Dandifile as seen on the archive; e.g., 'sub-123_ses-456+ecephys.nwb'.
    dandiset_id : six-digit string, "draft", or None
        The six-digit ID of the Dandiset.
    dandiset_version : string
        The specific published version of the Dandiset to inspect.
        If None, the latest version is used.
        If there are no published versions, then 'draft' is used instead.
    config : file path, dictionary, "dandi", or None, default: "dandi"
        If a file path, loads the dictionary configuration from the file.
        If a dictionary, it must be valid against the configuration schema.
        If "dandi", uses the requirements for DANDI validation.
    checks : list, optional
        list of checks to run
    ignore: list, optional
        Names of functions to skip.
    select: list, optional
        Names of functions to pick out of available checks.
    importance_threshold : string or Importance, optional
        Ignores tests with an assigned importance below this threshold.
        Importance has three levels:

            CRITICAL
                - potentially incorrect data
            BEST_PRACTICE_VIOLATION
                - very suboptimal data representation
            BEST_PRACTICE_SUGGESTION
                - improvable data representation

        The default is the lowest level, BEST_PRACTICE_SUGGESTION.
    skip_validate : bool, default: False
        Skip the PyNWB validation step.
        This may be desired for older NWBFiles (< schema version v2.10).
    client: dandi.dandiapi.DandiAPIClient
        The client object can be passed to avoid re-instantiation over an iteration.
    """
    if client is None:
        import dandi.dandiapi

        client = dandi.dandiapi.DandiAPIClient()

    dandiset = client.get_dandiset(dandiset_id=dandiset_id, version_id=dandiset_version)
    asset = dandiset.get_asset_by_path(path=dandi_file_path)
    asset_url = asset.get_content_url(follow_redirects=1, strip_query=True)

    for message in inspect_url(
        url=asset_url,
        config=config,
        checks=checks,
        ignore=ignore,
        select=select,
        importance_threshold=importance_threshold,
        skip_validate=skip_validate,
    ):
        message.file_path = dandi_file_path  # type: ignore
        yield message


def inspect_url(
    *,
    url: str,
    config: Union[str, pathlib.Path, dict, Literal["dandi"], None] = "dandi",
    checks: Union[list, None] = None,
    ignore: Union[list[str], None] = None,
    select: Union[list[str], None] = None,
    importance_threshold: Union[str, Importance] = Importance.BEST_PRACTICE_SUGGESTION,
    skip_validate: bool = False,
) -> Iterable[Union[InspectorMessage, None]]:
    """
    Inspect an explicit S3 URL.

    Parameters
    ----------
    url : string
        A URL referring to the cloud location of an NWB file.
        Commonly used with DANDI, where the URL has a form similar to:
        https://dandiarchive.s3.amazonaws.com/blobs/636/57e/63657e32-ad33-4625-b664-31699b5bf664

        Note: this must be the `https` URL, not the 's3://' form.
    config : file path, dictionary, "dandi", or None, default: "dandi"
        If a file path, loads the dictionary configuration from the file.
        If a dictionary, it must be valid against the configuration schema.
        If "dandi", uses the requirements for DANDI validation.
    checks : list, optional
        list of checks to run
    ignore: list, optional
        Names of functions to skip.
    select: list, optional
        Names of functions to pick out of available checks.
    importance_threshold : string or Importance, optional
        Ignores tests with an assigned importance below this threshold.
        Importance has three levels:

            CRITICAL
                - potentially incorrect data
            BEST_PRACTICE_VIOLATION
                - very suboptimal data representation
            BEST_PRACTICE_SUGGESTION
                - improvable data representation

        The default is the lowest level, BEST_PRACTICE_SUGGESTION.
    skip_validate : bool, default: False
        Whether to skip the PyNWB validation step.
    """
    import remfile

    filterwarnings(action="ignore", message="No cached namespaces found in .*")
    filterwarnings(action="ignore", message="Ignoring cached namespace .*")

    if isinstance(config, (str, pathlib.Path)):
        config = load_config(filepath_or_keyword=config)
    if isinstance(config, dict):
        validate_config(config=config)

    byte_stream = remfile.File(url=url)
    with (
        h5py.File(name=byte_stream) as file,
        pynwb.NWBHDF5IO(file=file) as io,
    ):
        if skip_validate is False:
            validation_errors = pynwb.validate(io=io)
            for validation_error in validation_errors:
                yield InspectorMessage(
                    message=validation_error.reason,
                    importance=Importance.PYNWB_VALIDATION,
                    check_function_name=validation_error.name,
                    location=validation_error.location,
                    file_path=url,
                )

        nwbfile = io.read()

        for message in inspect_nwbfile_object(
            nwbfile_object=nwbfile,
            config=config,
            checks=checks,
            ignore=ignore,
            select=select,
            importance_threshold=importance_threshold,
        ):
            message.file_path = url  # type: ignore
            yield message



"""Check functions specific to ImageSeries."""

from pathlib import Path, PurePosixPath, PureWindowsPath
from typing import Iterable, Optional

from pynwb.image import ImageSeries
from pynwb.ophys import TwoPhotonSeries

from .._registration import Importance, InspectorMessage, register_check
from ..tools import get_nwbfile_path_from_internal_object


@register_check(importance=Importance.CRITICAL, neurodata_type=ImageSeries)
def check_image_series_external_file_valid(image_series: ImageSeries) -> Optional[Iterable[InspectorMessage]]:
    """
    Check if the external_file specified by an ImageSeries actually exists.

    Best Practice: :ref:`best_practice_use_external_mode`
    """
    if image_series.external_file is None:
        return None
    nwbfile_path = Path(get_nwbfile_path_from_internal_object(neurodata_object=image_series))
    for file_path in image_series.external_file:
        file_path = file_path.decode() if isinstance(file_path, bytes) else file_path
        if not Path(file_path).is_absolute() and not (nwbfile_path.parent / file_path).exists():
            yield InspectorMessage(
                message=(
                    f"The external file '{file_path}' does not exist. Please confirm the relative location to the"
                    " NWBFile."
                )
            )

    return None


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=ImageSeries)
def check_image_series_external_file_relative(image_series: ImageSeries) -> Optional[Iterable[InspectorMessage]]:
    """
    Check if the external_file specified by an ImageSeries, if it exists, is relative.

    Best Practice: :ref:`best_practice_use_external_mode`
    """
    if image_series.external_file is None:
        return None
    for file_path in image_series.external_file:
        file_path = file_path.decode() if isinstance(file_path, bytes) else file_path
        if PureWindowsPath(file_path).is_absolute() or PurePosixPath(file_path).is_absolute():
            yield InspectorMessage(
                message=(
                    f"The external file '{file_path}' is not a relative path. "
                    "Please adjust the absolute path to be relative to the location of the NWBFile."
                )
            )

    return None


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=ImageSeries)
def check_image_series_data_size(image_series: ImageSeries, gb_lower_bound: float = 20.0) -> Optional[InspectorMessage]:
    """
    Check if an ImageSeries stored is larger than gb_lower_bound and suggests external file.

    Best Practice: :ref:`best_practice_use_external_mode`
    """
    # False positive case; TwoPhotonSeries are a subclass of ImageSeries, but it is very common and perfectly fine
    # to write lots of data using one without an external file
    if isinstance(image_series, TwoPhotonSeries):
        return None

    data = image_series.data

    if getattr(data, "compression", None) is not None:
        data_size_gb = data.id.get_storage_size() / 1e9
    else:
        data_size_gb = data.size * data.dtype.itemsize / 1e9

    if data_size_gb > gb_lower_bound:
        return InspectorMessage(message="ImageSeries is very large. Consider using external mode for better storage.")

    return None



"""Temporary module for thorough testing and evaluation of the proposed `read_nwbfile` helper function."""

from pathlib import Path
from typing import Literal, Optional, Union
from warnings import filterwarnings

import h5py
from hdmf.backends.io import HDMFIO
from hdmf_zarr import NWBZarrIO
from pynwb import NWBHDF5IO, NWBFile

BACKEND_IO_CLASSES = dict(
    hdf5=NWBHDF5IO,
    zarr=NWBZarrIO,
)


def _get_method(path: str) -> Literal["local", "fsspec"]:
    if path.startswith(("https://", "http://", "s3://")):
        return "fsspec"
    elif Path(path).exists():
        return "local"
    else:
        message = (
            f"Unable to automatically determine method. Path {path} does not appear to be a URL and is not a file on "
            f"the local filesystem."
        )
        raise ValueError(message)


def _init_fsspec(path: str) -> "fsspec.AbstractFileSystem":  # type: ignore
    import fsspec

    if path.startswith(("https://", "http://")):
        return fsspec.filesystem("http")
    elif path.startswith("s3://"):
        return fsspec.filesystem("s3", anon=True)
    else:
        message = f"Unable to initialize fsspec on path '{path}'."
        raise ValueError(message)


def _get_backend(path: str, method: Literal["local", "fsspec", "ros3"]) -> Union[str, Literal["hdf5", "zarr"]]:
    if method == "ros3":
        return "hdf5"

    possible_backends = []
    if method == "fsspec":
        filesystem = _init_fsspec(path=path)
        with filesystem.open(path=path, mode="rb") as file:
            for backend_name, backend_class in BACKEND_IO_CLASSES.items():
                if backend_class.can_read(path=file):
                    possible_backends.append(backend_name)
    else:
        for backend_name, backend_class in BACKEND_IO_CLASSES.items():
            if backend_class.can_read(path):
                possible_backends.append(backend_name)

    if len(possible_backends) > 1:  # pragma: no cover
        raise ValueError(
            f"More than one possible backend found - please choose from the following: {possible_backends}"
        )
    elif len(possible_backends) == 0:
        raise ValueError("No compatible backend found.")

    return possible_backends[0]


def read_nwbfile(
    nwbfile_path: Union[str, Path],
    method: Optional[Literal["local", "fsspec", "ros3"]] = None,
    backend: Optional[Literal["hdf5", "zarr"]] = None,
) -> NWBFile:
    """
    Read an NWB file using the specified (or auto-detected) method and specified (or auto-detected) backend.

    Parameters
    ----------
    nwbfile_path : str or pathlib.Path
        Path to the file on your system.
    method : "local", "fsspec", "ros3", or None (default)
        Where to read the file from; a local disk drive or steaming from an https:// or s3:// path.
        The default auto-detects based on the form of the path.
        When streaming, the default method is "fsspec".
        Note that "ros3" is specific to HDF5 backend files.
    backend : "hdf5", "zarr", or None (default)
        Type of backend used to write the file.
        The default auto-detects the type of the file.

    Returns
    -------
    nwbfile : pynwb.NWBFile
        The in-memory NWBFile object.
    """
    nwbfile, _ = _read_nwbfile_helper(nwbfile_path=nwbfile_path, method=method, backend=backend)

    # Note: do not be concerned about IO object closing due to garbage collection here
    # (the IO object is attached as an attribute to the NWBFile object)
    return nwbfile


def read_nwbfile_and_io(
    nwbfile_path: Union[str, Path],
    method: Optional[Literal["local", "fsspec", "ros3"]] = None,
    backend: Optional[Literal["hdf5", "zarr"]] = None,
) -> tuple[NWBFile, HDMFIO]:
    """
    Read an NWB file using the specified (or auto-detected) method and specified (or auto-detected) backend.

    Parameters
    ----------
    nwbfile_path : str or pathlib.Path
        Path to the file on your system.
    method : "local", "fsspec", "ros3", or None (default)
        Where to read the file from; a local disk drive or steaming from an https:// or s3:// path.
        The default auto-detects based on the form of the path.
        When streaming, the default method is "fsspec".
        Note that "ros3" is specific to HDF5 backend files.
    backend : "hdf5", "zarr", or None (default)
        Type of backend used to write the file.
        The default auto-detects the type of the file.

    Returns
    -------
    nwbfile : pynwb.NWBFile
        The in-memory NWBFile object.
    io : hdmf.backends.io.HDMFIO, optional
        Only passed if `return_io` is True.
        The initialized HDMFIO object used to read the file.
    """
    nwbfile, io = _read_nwbfile_helper(nwbfile_path=nwbfile_path, method=method, backend=backend)

    return nwbfile, io


def _read_nwbfile_helper(
    nwbfile_path: Union[str, Path],
    method: Optional[Literal["local", "fsspec", "ros3"]] = None,
    backend: Optional[Literal["hdf5", "zarr"]] = None,
) -> tuple[NWBFile, HDMFIO]:
    nwbfile_path = str(nwbfile_path)  # If pathlib.Path, cast to str; if already str, no harm done

    method = method or _get_method(nwbfile_path)
    if method != "local" and Path(nwbfile_path).exists():
        raise ValueError(
            f"The file ({nwbfile_path}) is a local path on your system, but the method ({method}) was selected! "
            "Please set method='local'."
        )
    if method == "local" and any(protocol in nwbfile_path for protocol in ["s3://", "https://"]):
        raise ValueError(
            f"The path ({nwbfile_path}) is an external URL, but the method (local) was selected! "
            "Please set method='fsspec' or 'ros3' (for HDF5 only)."
        )
    if method == "ros3" and nwbfile_path.startswith("s3://"):
        raise ValueError(
            "The ROS3 method was selected, but the URL starts with 's3://'! Please switch to an 'https://' URL."
        )

    chosen_backend = backend or _get_backend(path=nwbfile_path, method=method)
    # Temporary until .can_read() is able to work on streamed bytes
    if method == "local" and not BACKEND_IO_CLASSES[chosen_backend].can_read(path=nwbfile_path):
        raise IOError(
            f"The chosen backend ({chosen_backend}) is unable to read the file! Please select a different backend."
        )

    # Filter out some common warnings that don't really matter with `load_namespaces=True`
    filterwarnings(action="ignore", message="No cached namespaces found in .*")
    filterwarnings(action="ignore", message="Ignoring cached namespace .*")
    io_kwargs = dict(mode="r", load_namespaces=True)
    if method == "fsspec":
        fs = _init_fsspec(nwbfile_path)
        f = fs.open(nwbfile_path, "rb")
        file = h5py.File(f)
        io_kwargs.update(file=file)
    else:
        io_kwargs.update(path=nwbfile_path)
    if method == "ros3":
        io_kwargs.update(driver="ros3")
    io = BACKEND_IO_CLASSES[chosen_backend](**io_kwargs)
    nwbfile = io.read()

    return nwbfile, io



"""Primary functions for inspecting NWBFiles."""

import json
from collections.abc import Callable
from pathlib import Path
from types import FunctionType
from typing import Optional, Union

import jsonschema
import yaml

from ._registration import Importance, available_checks

INTERNAL_CONFIGS: dict[str, Path] = dict(
    dandi=Path(__file__).parent / "_internal_configs" / "dandi.inspector_config.yaml",
)


def validate_config(config: dict) -> None:
    """Validate an instance of configuration against the official schema."""
    config_schema_file_path = Path(__file__).parent / "_internal_configs" / "config.schema.json"
    with open(file=config_schema_file_path, mode="r") as fp:
        schema = json.load(fp=fp)
    jsonschema.validate(instance=config, schema=schema)


def _copy_function(function: Callable) -> Callable:
    """Copy the core parts of a given function, excluding wrappers, then return a new function."""
    copied_function = FunctionType(
        function.__code__, function.__globals__, function.__name__, function.__defaults__, function.__closure__
    )

    # in case f was given attrs (note this dict is a shallow copy)
    copied_function.__dict__.update(function.__dict__)

    return copied_function


def copy_check(check: Callable) -> Callable:
    """
    Copy a check function so that internal attributes can be adjusted without changing the original function.

    Required to ensure our configuration of functions in the registry does not effect the registry itself.

    Also copies the wrapper for auto-parsing results,
    see https://github.com/NeurodataWithoutBorders/nwbinspector/pull/218 for explanation.
    """
    copied_check = _copy_function(function=check)
    copied_check.__wrapped__ = _copy_function(function=check.__wrapped__)  # type: ignore

    return copied_check


def load_config(filepath_or_keyword: Union[str, Path]) -> dict:
    """
    Load a config dictionary either via keyword search of the internal configs, or an explicit filepath.

    Currently supported keywords are:
        - 'dandi'
            For all DANDI archive related practices, including validation and upload.
    """
    file = INTERNAL_CONFIGS.get(str(filepath_or_keyword), filepath_or_keyword)
    with open(file=file, mode="r") as stream:
        config = yaml.safe_load(stream=stream)

    return config


def configure_checks(
    checks: Optional[list] = None,
    config: Optional[dict] = None,
    ignore: Optional[list[str]] = None,
    select: Optional[list[str]] = None,
    importance_threshold: Importance = Importance.BEST_PRACTICE_SUGGESTION,
) -> list:
    """
    Filter a list of check functions (the entire base registry by default) according to the configuration.

    Parameters
    ----------
    checks : list of check functions, optional
        If None, defaults to all registered checks.
    config : dict
        Dictionary valid against our JSON configuration schema.
        Can specify a mapping of importance levels and list of check functions whose importance you wish to change.
        Typically loaded via json.load from a valid .json file
    ignore: list, optional
        Names of functions to skip.
    select: list, optional
        If loading all registered checks, this can be shorthand for selecting only a handful of them.
    importance_threshold : string, optional
        Ignores all tests with an post-configuration assigned importance below this threshold.
        Importance has three levels:

            CRITICAL
                - potentially incorrect data
            BEST_PRACTICE_VIOLATION
                - very suboptimal data representation
            BEST_PRACTICE_SUGGESTION
                - improvable data representation

        The default is the lowest level, BEST_PRACTICE_SUGGESTION.
    """
    checks = checks or available_checks

    if ignore is not None and select is not None:
        raise ValueError("Options 'ignore' and 'select' cannot both be used.")
    if importance_threshold not in Importance:
        raise ValueError(
            f"Indicated importance_threshold ({importance_threshold}) is not a valid importance level! Please choose "
            "from [CRITICAL_IMPORTANCE, BEST_PRACTICE_VIOLATION, BEST_PRACTICE_SUGGESTION]."
        )

    checks_out: list = []
    if config is not None:
        validate_config(config=config)
        ignore = ignore or []
        for check in checks:
            mapped_check = copy_check(check=check)
            for importance_name, func_names in config.items():
                if check.__name__ in func_names:
                    if importance_name == "SKIP":
                        ignore.append(check.__name__)
                        continue
                    mapped_check.importance = Importance[importance_name]  # type: ignore
                    # Output wrappers are apparently parsed at time of wrapping not of time of output return...
                    # Attempting to re-wrap the copied function if the importance level is being adjusted...
                    # From https://github.com/NeurodataWithoutBorders/nwbinspector/issues/302
                    # new_check_wrapper = _copy_function(function=mapped_check.__wrapped__)
                    # new_check_wrapper.importance = Importance[importance_name]
                    # mapped_check.__wrapped__ = new_check_wrapper
            checks_out.append(mapped_check)
    else:
        checks_out = checks
    if select:
        checks_out = [x for x in checks_out if x.__name__ in select]
    elif ignore:
        checks_out = [x for x in checks_out if x.__name__ not in ignore]
    if importance_threshold:
        checks_out = [x for x in checks_out if x.importance.value >= importance_threshold.value]  # type: ignore

    return checks_out



import warnings

message = "The 'register_checks' submodule has been deprecated. Please import the helper functions from the top-level package."

warnings.warn(message=message, category=DeprecationWarning, stacklevel=2)

# Still keep imports functional with warning for soft deprecation cycle
# TODO: remove after 9/15/2024
from .._types import InspectorMessage, Importance, Severity
from .._registration import register_check, available_checks



import warnings

message = (
    "All submodules below the level 'nwbinspector.checks'  have been deprecated. "
    "Please retrieve the specific check function(s) you wish to use from the `available_checks` registry "
    "at the top-level or import directly from 'nwbinspector.checks'."
)

warnings.warn(message=message, category=DeprecationWarning, stacklevel=2)

# Still keep imports functional with warning for soft deprecation cycle
# TODO: remove after 9/15/2024
from .._general import *



import warnings

message = (
    "The 'inspector_tools' submodule has been deprecated. "
    "Please import the helper functions from the top-level package."
)

warnings.warn(message=message, category=DeprecationWarning, stacklevel=2)

# Still keep imports functional with warning for soft deprecation cycle
# TODO: remove after 9/15/2024
from .._organization import organize_messages
from .._formatting import (
    format_messages,
    MessageFormatter,
    FormatterOptions,
    print_to_console,
    save_report,
    _get_report_header,
)



from ._utils import (
    get_data_shape,
    strtobool,
    format_byte_size,
    cache_data_selection,
    is_regular_series,
    is_ascending_series,
    is_dict_in_string,
    is_string_json_loadable,
    is_module_installed,
    get_package_version,
    calculate_number_of_cpu,
    get_data_shape,
    PathType,  # TODO: deprecate in favor of explicit typing
    FilePathType,  # TODO: deprecate in favor of explicit typing
    OptionalListOfStrings,  # TODO: deprecate in favor of explicit typing
)

__all__ = [
    "get_data_shape",
    "strtobool",
    "format_byte_size",
    "cache_data_selection",
    "is_regular_series",
    "is_ascending_series",
    "is_dict_in_string",
    "is_string_json_loadable",
    "is_module_installed",
    "get_package_version",
    "calculate_number_of_cpu",
    "get_data_shape",
]



"""Internally used tools specifically for rendering more human-readable output from collected check results."""

from enum import Enum
from typing import Optional

from natsort import natsorted

from ._registration import InspectorMessage


def _sort_unique_values(unique_values: list, reverse: bool = False) -> list:
    """Technically, the 'set' method applies basic sorting to the unique contents, but natsort is more general."""
    if any(unique_values) and isinstance(unique_values[0], Enum):
        return natsorted(unique_values, key=lambda x: -x.value, reverse=reverse)
    else:
        return natsorted(unique_values, reverse=reverse)


def organize_messages(
    messages: list[Optional[InspectorMessage]], levels: list[str], reverse: Optional[list[bool]] = None
) -> dict:
    """
    General function for organizing list of InspectorMessages.

    Returns a nested dictionary organized according to the order of the 'levels' argument.

    Parameters
    ----------
    messages : list of InspectorMessages
    levels : list of strings
        Each string in this list must correspond onto an attribute of the InspectorMessage class, excluding the
        'message' text and 'object_name' (this will be coupled to the 'object_type').
    reverse : list of bool, optional
        If provided, this should be a list of booleans that correspond to the 'levels' argument.
        If True, the values will be sorted in reverse order.
    """
    assert all([x not in levels for x in ["message", "object_name", "severity"]]), (
        "You must specify levels to organize by that correspond to attributes of the InspectorMessage class, excluding "
        "the text message, object_name, and severity."
    )
    if reverse is None:
        reverse = [False] * len(levels)
    unique_values = list(set(getattr(message, levels[0]) for message in messages))
    sorted_values = _sort_unique_values(unique_values, reverse=reverse[0])
    if len(levels) > 1:
        return {
            value: organize_messages(
                messages=[message for message in messages if getattr(message, levels[0]) == value],
                levels=levels[1:],
                reverse=reverse[1:],
            )
            for value in sorted_values
        }
    else:
        return {
            value: sorted(
                [message for message in messages if getattr(message, levels[0]) == value],
                key=lambda x: -x.severity.value,
            )
            for value in sorted_values
        }



"""Primary functions for inspecting NWBFiles."""

import importlib
import traceback
from collections import defaultdict
from concurrent.futures import ProcessPoolExecutor, as_completed
from pathlib import Path
from typing import Iterable, Optional, Type, Union
from warnings import filterwarnings, warn

import pynwb
from hdmf_zarr import ZarrIO
from natsort import natsorted
from tqdm import tqdm

from ._configuration import configure_checks
from ._registration import Importance, InspectorMessage, available_checks
from .tools._read_nwbfile import read_nwbfile, read_nwbfile_and_io
from .utils import (
    OptionalListOfStrings,
    PathType,
    calculate_number_of_cpu,
)


def inspect_all(
    path: PathType,
    config: Optional[dict] = None,
    ignore: OptionalListOfStrings = None,
    select: OptionalListOfStrings = None,
    importance_threshold: Union[str, Importance] = Importance.BEST_PRACTICE_SUGGESTION,
    n_jobs: int = 1,
    skip_validate: bool = False,
    progress_bar: bool = True,
    progress_bar_class: Type[tqdm] = tqdm,
    progress_bar_options: Optional[dict] = None,
    stream: bool = False,  # TODO: remove after 3/1/2025
    version_id: Optional[str] = None,  # TODO: remove after 3/1/2025
    modules: OptionalListOfStrings = None,
) -> Iterable[Union[InspectorMessage, None]]:
    """
    Inspect a local NWBFile or folder of NWBFiles and return suggestions for improvements according to best practices.

    Parameters
    ----------
    path : PathType
        File path to an NWBFile, folder path to iterate over recursively and scan all NWBFiles present, or a
        six-digit identifier of the DANDISet.
    config : dict, optional
        If a dictionary, it must be valid against our JSON configuration schema.
        Can specify a mapping of importance levels and list of check functions whose importance you wish to change.
        Typically loaded via json.load from a valid .json file
    ignore: list of strings, optional
        Names of functions to skip.
    select: list of strings, optional
        Names of functions to pick out of available checks.
    importance_threshold : string or Importance, optional
        Ignores tests with an assigned importance below this threshold.
        Importance has three levels:

            CRITICAL
                - potentially incorrect data
            BEST_PRACTICE_VIOLATION
                - very suboptimal data representation
            BEST_PRACTICE_SUGGESTION
                - improvable data representation

        The default is the lowest level, BEST_PRACTICE_SUGGESTION.
    n_jobs : int
        Number of jobs to use in parallel. Set to -1 to use all available resources.
        This may also be a negative integer x from -2 to -(number_of_cpus - 1) which acts like negative slicing by using
        all available CPUs minus x.
        Set to 1 (also the default) to disable.
    skip_validate : bool, optional
        Skip the PyNWB validation step.
        The default is False, which is recommended.
    progress_bar : bool, optional
        Display a progress bar while scanning NWBFiles.
        Defaults to True.
    progress_bar_class : type of tqdm.tqdm, optional
        The specific child class of tqdm.tqdm to use to make progress bars.
        Defaults to tqdm.tqdm, the most generic parent.
    progress_bar_options : dict, optional
        Dictionary of keyword arguments to pass directly to the progress_bar_class.
    modules : list of strings, optional
        List of external module names to load; examples would be namespace extensions.
        These modules may also contain their own custom checks for their extensions.
    """
    in_path = Path(path)
    importance_threshold = (
        Importance[importance_threshold] if isinstance(importance_threshold, str) else importance_threshold
    )
    modules = modules or []

    for module in modules:
        importlib.import_module(module)

    # TODO: remove these blocks after 3/1/2025
    if version_id is not None:
        deprecation_message = (
            "The `version_id` argument is deprecated and will be removed after 3/1/2025. "
            "Please call `nwbinspector.inspect_dandiset` with the argument `dandiset_version` instead."
        )
        warn(message=deprecation_message, category=DeprecationWarning, stacklevel=2)
    if stream:
        from ._dandi_inspection import inspect_dandiset

        warning_message = (
            "The `stream` argument is deprecated and will be removed after 3/1/2025. "
            "Please call `nwbinspector.inspect_dandiset` instead."
        )
        warn(message=warning_message, category=DeprecationWarning, stacklevel=2)

        for message in inspect_dandiset(
            dandiset_id=str(path),
            dandiset_version=version_id,
            config=config,
            ignore=ignore,
            select=select,
            skip_validate=skip_validate,
        ):
            yield message

        return None

    calculated_number_of_jobs = calculate_number_of_cpu(requested_cpu=n_jobs)
    if progress_bar_options is None:
        progress_bar_options = dict(position=0, leave=False)

    if in_path.is_dir() and (in_path.match("*.nwb*")) and ZarrIO.can_read(in_path):
        nwbfiles = [in_path]  # if it is a zarr directory
    elif in_path.is_dir():
        nwbfiles = list(in_path.rglob("*.nwb*"))

        # Remove any macOS sidecar files
        nwbfiles = [nwbfile for nwbfile in nwbfiles if not nwbfile.name.startswith("._")]
    elif in_path.is_file():
        nwbfiles = [in_path]
    else:
        raise ValueError(f"{in_path} should be a directory or an NWB file.")
    # Filtering of checks should apply after external modules are imported, in case those modules have their own checks
    checks = configure_checks(config=config, ignore=ignore, select=select, importance_threshold=importance_threshold)

    # Manual identifier check over all files in the folder path
    identifiers = defaultdict(list)
    for nwbfile_path in nwbfiles:
        try:
            nwbfile = read_nwbfile(nwbfile_path=nwbfile_path)
            identifiers[nwbfile.identifier].append(nwbfile_path)
        except Exception as exception:
            continue  # read failure errors will be returned as part of inspect_nwbfile

    if len(identifiers) != len(nwbfiles):
        for identifier, nwbfiles_with_identifier in identifiers.items():
            if len(nwbfiles_with_identifier) > 1:
                yield InspectorMessage(
                    message=(
                        f"The identifier '{identifier}' is used across the .nwb files: "
                        f"{natsorted([x.name for x in nwbfiles_with_identifier])}. "
                        "The identifier of any NWBFile should be a completely unique value - "
                        "we recommend using uuid4 to achieve this."
                    ),
                    importance=Importance.CRITICAL,
                    check_function_name="check_unique_identifiers",
                    object_type="NWBFile",
                    object_name="root",
                    location="/",
                    file_path=str(path),
                )

    nwbfiles_iterable = nwbfiles
    if progress_bar:
        nwbfiles_iterable = progress_bar_class(nwbfiles_iterable, **progress_bar_options)
    if calculated_number_of_jobs == 1:
        for nwbfile_path in nwbfiles_iterable:  # type: ignore
            for message in inspect_nwbfile(nwbfile_path=nwbfile_path, checks=checks, skip_validate=skip_validate):
                yield message
    else:
        progress_bar_options.update(total=len(nwbfiles))
        futures = []
        # concurrents uses None instead of -1 for 'auto' mode
        max_workers = None if calculated_number_of_jobs == -1 else calculated_number_of_jobs
        with ProcessPoolExecutor(max_workers=max_workers) as executor:
            for nwbfile_path in nwbfiles:
                futures.append(
                    executor.submit(
                        _pickle_inspect_nwb,
                        nwbfile_path=str(nwbfile_path),
                        checks=checks,
                        skip_validate=skip_validate,
                    )
                )
            async_nwbfiles_iterable = as_completed(futures)
            if progress_bar:
                async_nwbfiles_iterable = progress_bar_class(async_nwbfiles_iterable, **progress_bar_options)
            for future in async_nwbfiles_iterable:
                for message in future.result():
                    if stream:
                        message.file_path = nwbfiles[message.file_path]
                    yield message


def _pickle_inspect_nwb(
    nwbfile_path: str,
    checks: Optional[list] = None,
    skip_validate: bool = False,
) -> Iterable[Union[InspectorMessage, None]]:
    """Auxiliary function for inspect_all to run in parallel using the ProcessPoolExecutor."""
    checks = checks or available_checks

    return list(inspect_nwbfile(nwbfile_path=nwbfile_path, checks=checks, skip_validate=skip_validate))


def inspect_nwbfile(
    nwbfile_path: Union[str, Path],
    driver: Optional[str] = None,  # TODO: remove after 3/1/2025
    skip_validate: bool = False,
    max_retries: Optional[int] = None,  # TODO: remove after 3/1/2025
    checks: Optional[list] = None,
    config: Optional[dict] = None,
    ignore: OptionalListOfStrings = None,
    select: OptionalListOfStrings = None,
    importance_threshold: Union[str, Importance] = Importance.BEST_PRACTICE_SUGGESTION,
) -> Iterable[Union[InspectorMessage, None]]:
    """
    Open an NWB file, inspect the contents, and return suggestions for improvements according to best practices.

    Parameters
    ----------
    nwbfile_path : FilePathType
        Path to the NWB file on disk or on S3.
    skip_validate : bool
        Skip the PyNWB validation step.
        The default is False, which is recommended.
    checks : list, optional
        List of checks to run.
    config : dict
        Dictionary valid against our JSON configuration schema.
        Can specify a mapping of importance levels and list of check functions whose importance you wish to change.
        Typically loaded via `json.load` from a valid .json file.
    ignore: list, optional
        Names of functions to skip.
    select: list, optional
        Names of functions to pick out of available checks.
    importance_threshold : string or Importance, optional
        Ignores tests with an assigned importance below this threshold.
        Importance has three levels:

            CRITICAL
                - potentially incorrect data
            BEST_PRACTICE_VIOLATION
                - very suboptimal data representation
            BEST_PRACTICE_SUGGESTION
                - improvable data representation

        The default is the lowest level, BEST_PRACTICE_SUGGESTION.
    """
    checks = checks or available_checks
    # TODO: remove error after 3/1/2025
    if driver is not None or max_retries is not None:
        message = (
            "The `driver` and `max_retries` arguments are deprecated and will be removed after 3/1/2025. "
            "Please call `nwbinspector.inspect_dandi_file_path` instead."
        )
        raise ValueError(message)

    nwbfile_path = str(nwbfile_path)
    filterwarnings(action="ignore", message="No cached namespaces found in .*")
    filterwarnings(action="ignore", message="Ignoring cached namespace .*")

    io = None
    try:
        in_memory_nwbfile, io = read_nwbfile_and_io(nwbfile_path=nwbfile_path)

        if not skip_validate:
            validation_result = pynwb.validate(path=nwbfile_path)
            if isinstance(validation_result, tuple):
                validation_errors = validation_result[0]
            else:
                validation_errors = validation_result

            for validation_error in validation_errors:
                yield InspectorMessage(
                    message=validation_error.reason,
                    importance=Importance.PYNWB_VALIDATION,
                    check_function_name=validation_error.name,
                    location=validation_error.location,
                    file_path=nwbfile_path,
                )

        for inspector_message in inspect_nwbfile_object(
            nwbfile_object=in_memory_nwbfile,
            checks=checks,
            config=config,
            ignore=ignore,
            select=select,
            importance_threshold=importance_threshold,
        ):
            inspector_message.file_path = nwbfile_path  # type: ignore
            yield inspector_message
    except Exception as exception:
        exception_name = f"{type(exception).__module__}.{type(exception).__name__}"
        yield InspectorMessage(
            message=traceback.format_exc(),
            importance=Importance.ERROR,
            check_function_name=(
                f"During io.read(), an error occurred: {exception_name}. "
                f"This indicates that PyNWB was unable to read the file. "
                f"See the traceback message for more details."
            ),
            file_path=nwbfile_path,
        )
    finally:
        if io is not None:
            io.close()  # close the io object in case of exceptions or when inspection is complete


# TODO: deprecate once subject types and dandi schemas have been extended
def _intercept_in_vitro_protein(nwbfile_object: pynwb.NWBFile, checks: Optional[list] = None) -> list:
    """
    If the special 'protein' subject_id is specified, return a truncated list of checks to run.

    This is a temporary method for allowing upload of certain in vitro data to DANDI and
    is expected to be replaced in future versions.
    """
    checks = checks or available_checks

    subject_related_check_names = [
        "check_subject_exists",
        "check_subject_id_exists",
        "check_subject_sex",
        "check_subject_species_exists",
        "check_subject_species_form",
        "check_subject_age",
        "check_subject_proper_age_range",
    ]
    subject_related_dandi_requirements = [
        check.importance == Importance.CRITICAL  # type: ignore
        for check in checks
        if check.__name__ in subject_related_check_names
    ]

    subject = getattr(nwbfile_object, "subject", None)
    if (
        any(subject_related_dandi_requirements)
        and subject is not None
        and (getattr(subject, "subject_id") or "").startswith("protein")
    ):
        non_subject_checks = [check for check in checks if check.__name__ not in subject_related_check_names]
        return non_subject_checks

    return checks


def inspect_nwbfile_object(
    nwbfile_object: pynwb.NWBFile,
    checks: Optional[list] = None,
    config: Optional[dict] = None,
    ignore: Optional[list[str]] = None,
    select: Optional[list[str]] = None,
    importance_threshold: Union[str, Importance] = Importance.BEST_PRACTICE_SUGGESTION,
) -> Iterable[Union[InspectorMessage, None]]:
    """
    Inspect an in-memory NWBFile object and return suggestions for improvements according to best practices.

    Parameters
    ----------
    nwbfile_object : NWBFile
        An in-memory NWBFile object.
    checks : list, optional
        list of checks to run
    config : dict, optional
        Dictionary valid against our JSON configuration schema.
        Can specify a mapping of importance levels and list of check functions whose importance you wish to change.
        Typically loaded via json.load from a valid .json file
    ignore: list, optional
        Names of functions to skip.
    select: list, optional
        Names of functions to pick out of available checks.
    importance_threshold : string or Importance, optional
        Ignores tests with an assigned importance below this threshold.
        Importance has three levels:

            CRITICAL
                - potentially incorrect data
            BEST_PRACTICE_VIOLATION
                - very suboptimal data representation
            BEST_PRACTICE_SUGGESTION
                - improvable data representation

        The default is the lowest level, BEST_PRACTICE_SUGGESTION.
    """
    checks = checks or available_checks
    importance_threshold = (
        Importance[importance_threshold] if isinstance(importance_threshold, str) else importance_threshold
    )
    if any(argument is not None for argument in [config, ignore, select, importance_threshold]):
        checks = configure_checks(
            checks=checks, config=config, ignore=ignore, select=select, importance_threshold=importance_threshold
        )

    subject_dependent_checks = _intercept_in_vitro_protein(nwbfile_object=nwbfile_object, checks=checks)

    for inspector_message in run_checks(nwbfile=nwbfile_object, checks=subject_dependent_checks):
        yield inspector_message


def run_checks(
    nwbfile: pynwb.NWBFile,
    checks: list,
    progress_bar_class: Optional[Type[tqdm]] = None,
    progress_bar_options: Optional[dict] = None,
) -> Iterable[Union[InspectorMessage, None]]:
    """
    Run checks on an open NWBFile object.

    Parameters
    ----------
    nwbfile : pynwb.NWBFile
        The in-memory pynwb.NWBFile object to run the checks on.
    checks : list of check functions
        The list of check functions that will be run on the in-memory pynwb.NWBFile object.
    progress_bar_class : type of tqdm.tqdm, optional
        The specific child class of tqdm.tqdm to use to make progress bars.
        Defaults to not displaying progress per set of checks over an individual file.
    progress_bar_options : dict, optional
        Dictionary of keyword arguments to pass directly to the `progress_bar_class`.

    Yields
    ------
    results : a generator of InspectorMessage objects
        A generator that returns a message on each iteration, if any are triggered by downstream conditions.
        Otherwise, has length zero (if cast as `list`), or raises `StopIteration` (if explicitly calling `next`).
    """
    if progress_bar_class is not None:
        check_progress = progress_bar_class(iterable=checks, total=len(checks), **progress_bar_options)
    else:
        check_progress = checks

    for check_function in check_progress:
        for nwbfile_object in nwbfile.objects.values():
            if check_function.neurodata_type is not None and not issubclass(
                type(nwbfile_object), check_function.neurodata_type
            ):
                continue

            try:
                output = check_function(nwbfile_object)
            # if an individual check fails, include it in the report and continue with the inspection
            except Exception as exception:
                check_function_name = (
                    f"During evaluation of '{check_function.__name__}' - {type(exception)}: {str(exception)}"
                )
                output = InspectorMessage(
                    message=traceback.format_exc(),
                    importance=Importance.ERROR,
                    check_function_name=check_function_name,
                    file_path="unknown",
                )
            if isinstance(output, InspectorMessage):
                # temporary solution to https://github.com/dandi/dandi-cli/issues/1031
                if output.importance != Importance.ERROR:
                    output.importance = check_function.importance
                yield output
            elif output is not None:
                for x in output:
                    x.importance = check_function.importance
                    yield x



import warnings

message = (
    "The 'nwbinspector.nwbinspector' submodule has been deprecated. "
    "Please import the helper functions from the top-level package."
)

warnings.warn(message=message, category=DeprecationWarning, stacklevel=2)

# Still keep imports functional with warning for soft deprecation cycle
# TODO: remove after 9/15/2024
from .._configuration import (
    INTERNAL_CONFIGS,
    validate_config,
    copy_check,
    load_config,
    configure_checks,
)
from .._nwb_inspection import (
    inspect_all,
    inspect_nwbfile,
    inspect_nwbfile_object,
    run_checks,
)
from .._formatting import InspectorOutputJSONEncoder



"""Checks specific to the Images neurodata type."""

from typing import Optional

from pynwb.base import Images
from pynwb.image import IndexSeries

from .._registration import Importance, InspectorMessage, register_check


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=Images)
def check_order_of_images_unique(images: Images) -> Optional[InspectorMessage]:
    """Check that all the values in the order_of_images field of an Images object are unique."""
    if images.order_of_images is None:
        return None
    if not len(set(images.order_of_images)) == len(images.order_of_images):
        return InspectorMessage(message="order_of_images should have unique values.")

    return None


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=Images)
def check_order_of_images_len(images: Images) -> Optional[InspectorMessage]:
    """Check that all the values in the order_of_images field of an Images object are unique."""
    if images.order_of_images is None:
        return None
    if not len(images.order_of_images) == len(images.images):
        return InspectorMessage(
            message=f"Length of order_of_images ({len(images.order_of_images)}) does not match the number of "
            f"images ({len(images.images)})."
        )

    return None


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=IndexSeries)
def check_index_series_points_to_image(index_series: IndexSeries) -> Optional[InspectorMessage]:
    if index_series.indexed_timeseries is not None:
        return InspectorMessage(
            message="Pointing an IndexSeries to a TimeSeries is deprecated. Please point to an Images "
            "container instead."
        )

    return None



"""Check functions that can apply to any object inside an NWBContainer."""

import os
from typing import Iterable, Optional

import h5py
import zarr
from pynwb import NWBContainer

from .._registration import Importance, InspectorMessage, Severity, register_check


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=NWBContainer)
def check_large_dataset_compression(
    nwb_container: NWBContainer, gb_lower_bound: float = 20.0
) -> Optional[InspectorMessage]:
    """
    If the data in the Container object is a 'large' h5py.Dataset, check if it has compression enabled.

    Will only return an inspector warning if the size of the h5py.Dataset is larger than the
    gb_lower_bound (default of 20 GB).

    Best Practice: :ref:`best_practice_compression`
    """
    for field in getattr(nwb_container, "fields", dict()).values():
        if not isinstance(field, (h5py.Dataset, zarr.Array)):
            continue

        compression_indicator = None
        if isinstance(field, h5py.Dataset):
            compression_indicator = field.compression
        elif isinstance(field, zarr.Array):
            compression_indicator = field.compressor

        field_size_bytes = field.size * field.dtype.itemsize
        if compression_indicator is None and field_size_bytes > gb_lower_bound * 1e9:
            return InspectorMessage(
                severity=Severity.HIGH,
                message=f"{os.path.split(field.name)[1]} is a large uncompressed dataset! Please enable compression.",
            )

    return None


@register_check(importance=Importance.BEST_PRACTICE_SUGGESTION, neurodata_type=NWBContainer)
def check_small_dataset_compression(
    nwb_container: NWBContainer,
    gb_severity_threshold: float = 10.0,
    mb_lower_bound: float = 50.0,
    gb_upper_bound: float = 20.0,  # 20 GB upper bound to prevent double-raise
) -> Optional[InspectorMessage]:
    """
    If the data in the Container object is a h5py.Dataset, check if it has compression enabled.

    Will only return an inspector warning if the size of the h5py.Dataset is larger than mb_lower_bound (default 50 MB)
    and smaller than gb_upper_bound (default of 20 GB).

    Best Practice: :ref:`best_practice_compression`
    """
    for field in getattr(nwb_container, "fields", dict()).values():
        if not isinstance(field, (h5py.Dataset, zarr.Array)):
            continue

        compression_indicator = None
        if isinstance(field, h5py.Dataset):
            compression_indicator = field.compression
        elif isinstance(field, zarr.Array):
            compression_indicator = field.compressor

        if (
            compression_indicator is None
            and mb_lower_bound * 1e6 < field.size * field.dtype.itemsize < gb_upper_bound * 1e9
        ):
            if field.size * field.dtype.itemsize > gb_severity_threshold * 1e9:
                severity = Severity.HIGH
            else:
                severity = Severity.LOW
            return InspectorMessage(
                severity=severity,
                message=(
                    f"{os.path.split(field.name)[1]} is not compressed. Consider enabling compression when writing a "
                    "dataset."
                ),
            )

    return None


@register_check(importance=Importance.BEST_PRACTICE_SUGGESTION, neurodata_type=NWBContainer)
def check_empty_string_for_optional_attribute(nwb_container: NWBContainer) -> Optional[Iterable[InspectorMessage]]:
    """
    Check if any NWBContainer has optional fields that are written as an empty string.

    These values should just be omitted instead.

    Parameters
    ----------
    nwb_container: NWBContainer

    Best Practice: :ref:`best_practice_placeholders`
    """
    docval_args = type(nwb_container).__init__.__docval__["args"]
    optional_attrs = [
        arg["name"] for arg in docval_args if arg["type"] is str and "default" in arg and arg["default"] is None
    ]
    fields = [attr for attr in optional_attrs if getattr(nwb_container, attr) == ""]
    for field in fields:
        yield InspectorMessage(
            message=f'The attribute "{field}" is optional and you have supplied an empty string. Improve my omitting '
            "this attribute (in MatNWB or PyNWB) or entering as None (in PyNWB)"
        )

    return None
